# -*- coding: utf-8 -*-
"""
Summary
-------

Models built around the :class:`~radis.spectrum.spectrum.Spectrum` class


Routine Listing
---------------

- :func:`~radis.spectrum.models.Transmittance`
- :func:`~radis.spectrum.models.Radiance`


-------------------------------------------------------------------------------


"""

from __future__ import print_function, absolute_import, division, unicode_literals
from radis.spectrum.spectrum import Spectrum
import numpy as np


# %% Array-to-Spectrum functions


def calculated_spectrum(w, I, wunit='nm', Iunit='mW/cm2/sr/nm',
                        conditions=None, cond_units=None, populations=None,
                        name=None): # -> Spectrum:
    ''' Convert (w, I) into a Spectrum object that has unit conversion, plotting
    and slit convolution capabilities


    Parameters    
    ----------

    w, I: np.array
        wavelength and intensity

    wunit: 'nm', 'cm-1'
        wavespace unit

    Iunit: str
        intensity unit (can be 'counts', 'mW/cm2/sr/nm', etc...). Default
        'mW/cm2/sr/nm' (note that non-convoluted Specair spectra are in 'mW/cm2/sr/µm')


    Other Parameters
    ----------------

    conditions: dict
        (optional) calculation conditions to be stored with Spectrum. Default ``None``

    cond_units: dict
        (optional) calculation conditions units. Default ``None``

    populations: dict
        populations to be stored in Spectrum. Default ``None``

    name: str
        (optional) give a name


    See Also
    --------

    :func:`~radis.spectrum.spectrum.transmittance_spectrum`, 
    :func:`~radis.spectrum.spectrum.experimental_spectrum`,
    :meth:`~radis.spectrum.spectrum.Spectrum.from_array`,
    :meth:`~radis.spectrum.spectrum.Spectrum.from_txt`,
    :func:`~radis.tools.database.load_spec`

    '''

    return Spectrum.from_array(np.array(w), np.array(I), 'radiance_noslit',
                               waveunit=wunit, unit=Iunit,
                               conditions=conditions, cond_units=cond_units,
                               populations=populations,
                               name=name)


def transmittance_spectrum(w, T, wunit='nm', Tunit='I/I0',
                           conditions=None, cond_units=None,
                           name=None): # -> Spectrum:
    ''' Convert (w, I) into a Spectrum object that has unit conversion, plotting
    and slit convolution capabilities


    Parameters    
    ----------

    w, I: np.array
        wavelength and transmittance (no slit)

    wunit: 'nm', 'cm-1'
        wavespace unit

    Iunit: str
        intensity unit. Default 'I/I0'


    Other Parameters
    ----------------

    conditions: dict
        (optional) calculation conditions to be stored with Spectrum

    cond_units: dict
        (optional) calculation conditions units

    name: str
        (optional) give a name


    See Also
    --------

    :func:`~radis.spectrum.spectrum.calculated_spectrum`, 
    :func:`~radis.spectrum.spectrum.experimental_spectrum`,
    :meth:`~radis.spectrum.spectrum.Spectrum.from_array`,
    :meth:`~radis.spectrum.spectrum.Spectrum.from_txt`,
    :func:`~radis.tools.database.load_spec`

    '''

    return Spectrum.from_array(np.array(w), np.array(T), 'transmittance_noslit',
                               waveunit=wunit, unit=Tunit,
                               conditions=conditions, cond_units=cond_units,
                               name=name)


def experimental_spectrum(w, I, wunit='nm', Iunit='counts',
                          conditions=None, cond_units=None, name=None): # -> Spectrum:
    ''' Convert (w, I) into a Spectrum object that has unit conversion and plotting
    capabilities. Convolution is not available as the spectrum is assumed to
    be measured experimentally (hence deconvolution of the slit function would
    be required)


    Parameters    
    ----------

    w, I: np.array
        wavelength and intensity

    wunit: 'nm', 'cm-1'
        wavespace unit

    Iunit: str
        intensity unit (can be 'counts', 'mW/cm2/sr/nm', etc...). Default
        'counts' (default Winspec output)

    Other Parameters
    ----------------

    conditions: dict
        (optional) calculation conditions to be stored with Spectrum

    cond_units: dict
        (optional) calculation conditions units

    name: str
        (optional) give a name


    See Also
    --------

    :func:`~radis.spectrum.spectrum.calculated_spectrum`, 
    :func:`~radis.spectrum.spectrum.transmittance_spectrum`, 
    :meth:`~radis.spectrum.spectrum.Spectrum.from_array`,
    :meth:`~radis.spectrum.spectrum.Spectrum.from_txt`,
    :func:`~radis.tools.database.load_spec`

    '''

    return Spectrum.from_array(np.array(w), np.array(I), 'radiance',
                               waveunit=wunit, unit=Iunit,
                               conditions=conditions, cond_units=cond_units,
                               name=name)


# -*- coding: utf-8 -*-
"""
Variants of default Partition function tabulators and calculators, 
based on CDSD-4000 format, specifically for CO2

Routine Listing
---------------

Partition functions:

- :class:`~radis.levels.partfunc_cdsd.PartFuncCO2_CDSDcalc`
- :class:`~radis.levels.partfunc_cdsd.PartFuncCO2_CDSDtab`

Which inherit from:

- :class:`~radis.levels.partfunc.RovibParFuncCalculator`
- :class:`~radis.levels.partfunc.RovibParFuncTabulator`


-------------------------------------------------------------------------------


"""

from __future__ import print_function, absolute_import, division, unicode_literals

from scipy.interpolate import splrep, splev
import pandas as pd
import radis
from radis import OLDEST_COMPATIBLE_VERSION
from radis.db.molecules import ElectronicState
from radis.misc.cache_files import load_h5_cache_file
from radis.misc.cache_files import filter_metadata, save_to_hdf
from radis.lbl.labels import (vib_lvl_name_cdsd_p, vib_lvl_name_cdsd_pc,
                             vib_lvl_name_cdsd_pcN, vib_lvl_name_cdsd_pcJN)
from radis.levels.partfunc import RovibParFuncCalculator, RovibParFuncTabulator
from warnings import warn
from os.path import exists


# %% Variants of Tabulated partition functions (interpolate)

class PartFuncCO2_CDSDtab(RovibParFuncTabulator):
    ''' Return partition function of CO2 using a spline interpolation of
    tabulated values

    Parameters
    ----------

    T: temperature (K)
        gas temperature (equilibrium)

    Notes
    -----

    Partition function calculated in CDSD by direct summation (Jmax=300)

    '''

    def __init__(self, isotope, database):
        ''' Get partition function for one isotope only

        (note that this forces to reload the file once per isotope,
        but at least we have a clean layout with one object
        per isotope)

        '''

        Itable = {1: '12C16O2',
                  2: '13C16O2',
                  3: '16O12C18O',
                  4: '16O12C17O'}

        # Check input
        if not isotope in Itable:
            raise KeyError('Isotope {0} not defined in CDSD tabulated '.format(isotope) +
                           'partition functions. Only the following are: {0}'.format(
                list(Itable.keys())))

        # Read partition function tabulated data
        parsum = pd.read_csv(database, comment='#', delim_whitespace=True)
        if not 'T(K)' in list(parsum.keys()):
            raise KeyError(
                'Missing columns ({0}) in {1}'.format('T(K)', database))

        # Define spline interpolation
        isoname = Itable[isotope]
        tck = splrep(parsum['T(K)'], parsum[isoname])

        self.molecule = 2    # 'CO2'
        self.iso = isotope

        self.tck = tck   # dictionary
        self.Tmin = parsum['T(K)'].min()
        self.Tmax = parsum['T(K)'].max()

    def _inrange(self, T):
        ''' Allow for 5% extrapolation (ex: 296K / 300K) ) '''
        return ((self.Tmin*0.95 <= T).all() and (self.Tmax*1.05 >= T).all())

    def _at(self, T):
        ''' Get partition function at temperature T

        Called by :meth:`radis.levels.partfunc.RovibParFuncTabulator.at`
        '''
        try:
            assert(self._inrange(T))
        except:
            raise ValueError(
                'Temperature: {0} is out of bounds {1}-{2}'.format(T, self.Tmin, self.Tmax))

        return splev(T, self.tck)





# %% Calculated partition functions (either from energy levels, or ab initio)

class PartFuncCO2_CDSDcalc(RovibParFuncCalculator):
    ''' Calculate Partition Function from energy levels (and maybe export
    a tabulated database). Note: normally, ZPE (zero point energy) has
    already been taken into account in both the transition database and the
    energy database.

    Parameters
    ----------

    energy_levels: filename
        path to energy levels (to calculate Partition Function) for ``isotope``

    isotope: int
        which isotope we're dealing with. Default 1

    levelsfmt: ``'cdsd-p'``, ``'cdsd-pc'``, ``'cdsd-pcN'``, ``'cdsd-hamil'``, or ``None``
        the format of the Energy Database, and in particular how ``Evib`` and ``Erot``
        have been calculated. A vibrational level in the CDSD (p,c,J,N) nomenclature
        can be defined for levels that share a same (p), (p,c) or (p,c,N), where 
        ``p`` is the polyad number, ``c`` is the Wang symmetry number, and ``N`` 
        is the ranking index of a (p,c,J) group. Default ``'cdsd-pc'``.
        
        If ``None``, dont label the levels. Wont be able to use the EnergyDatabase to fetch
        vibrational energies for lines, however it can still be used to 
        calculate Partition functions independently from a Spectrum calculation

    Other Parameters
    ----------------
    
    use_cached: ``True``, ``False``, or ``'regen'``, ``'force'``
        if ``True``, use (and generate if doesnt exist) a ``.h5`` file.
        If ``'regen'``, regenerate cache file. If ``'force'``, raise an error
        if file doesnt exist. Default ``True``

    use_json: boolean
        deprecated. Better use h5 now.

    Notes
    -----

    Database format:

    Taskhkun database updated with ranking number (n) & total rank (N) of
    block, Evib and Erot (cm-1)  and jref

    jref is the lowest rotational number for which Evib was calculated.
    It should be 0 (so that Evib(p, c, j, n) = E(p, c, 0, n) but sometimes
    a j=0 state is not present for a given (p, c, n) state. So Evib is
    calculated using j=1 or j=2. It creates a small error!

    This could be improved extrapolating the j=0 energy plotting all
    E(p, c, j, n) = f_pcn(j)

    Example of table format::

        # calculated rovibrational energy levels of 12C16O2
        # =================================================
        # S.Tashkun, Zuev Institute of Atmospheric Optics, Tomsk, Russia
        # date: 17.03.2017
        #
        # zero point energy ZPE (cm-1) =  2531.828
        #
        # p = 2v1 + v2 + 3v3 - polyad number
        # j - rotational quantum number
        # c - Wang symmetry (1-'e'; 2-'f')
        # N - ranking number of energy levels of (p,j,c) blocks
        # n - total rank of (p,j,c) blocks
        #
        # Calculation limitations:
        # pmax = 40
        # jmax = 300
        # Ecut = 44600 cm-1
        # ---------------
        p	c	j	N	n	E	Evib	Erot	jref
        0	1	0	1	1	0.000	0.000	0.000	0
        0	1	2	1	1	2.341	0.000	2.341	0
        0	1	4	1	1	7.804	0.000	7.804	0
        0	1	6	1	1	16.389	0.000	16.389	0
        0	1	8	1	1	28.095	0.000	28.095	0


    '''

    def __init__(self, energy_levels, isotope, levelsfmt, #='cdsd-pc', 
                 use_cached=True, use_json=None, verbose=True):

        # %% Init

        # Initialize PartitionFunctionCalculator for this electronic state
        ElecState = ElectronicState('CO2', isotope, 'X', '1Σu+')
        super(PartFuncCO2_CDSDcalc, self).__init__(ElecState)

        # Check inputs ('return' is not mentionned in signature. it will just return
        # after cache name is given)
        assert use_cached in [True, False, 'regen', 'force', 'return']
        if isotope not in [1, 2]:
            raise ValueError(
                'CDSD Energies not defined for isotope: {0}'.format(isotope))
        if use_json is not None:
            warn(DeprecationWarning(
                'use_json replaced with faster HDF5-based use_cached'))
        # Get vibrational level definitions that match Energy Database (in particular
        # how Evib and Erot are calculated)
        # This is needed to be able to match the levels in the Line Database and
        # the levels in the Energy database
        if levelsfmt == 'cdsd-p':
            viblvl_label = 'p'
        elif levelsfmt == 'cdsd-pc':
            viblvl_label = 'pc'
        elif levelsfmt == 'cdsd-pcN':
            viblvl_label = 'pcN'
        elif levelsfmt == 'cdsd-hamil':
            viblvl_label = 'pcJN'
        elif levelsfmt is None:
            # dont label the levels. Wont be able to use the EnergyDatabase to fetch
            # vibrational energies for lines, however it can still be used to 
            # calculate Partition functions independently from a Spectrum calculation
            viblvl_label = None
        else:
            raise ValueError('Unknown Energy database format: levelsfmt = `{0}`'.format(
                             levelsfmt)+'. Use one of: `cdsd-p`, `cdsd-pc`, `cdsd-pcN`,`cdsd-hamil`')

        # Store defaults
        self.verbose = verbose
        self.use_cached = use_cached
        self.levelsfmt = levelsfmt
        self.viblvl_label = viblvl_label

        # Get variables to store in metadata  (after default values have been set)
        molecule = 'CO2'                # will be stored in cache file metadata
        _discard = ['self', 'energy_levels', 'verbose', 'ElecState', 'electronic_state',
                    'use_json', 'use_cached']
        # (dev) locals() automatically stores all variables: levelsfmt, viblvl_label, etc.
        metadata = filter_metadata(locals(), discard_variables=_discard)

        # %% Get levels

        # Function of use_cached value:
        # ... if True, use (and generate if doesnt exist) cache file.
        # ... if 'regen', regenerate cache file. If 'force', raise an error
        # ... if file doesnt exist.
        # If file is deprecated, regenerate it unless 'force' was used

        # Load cache file if exists

        cachefile = energy_levels+'.h5'
        self.cachefile = cachefile

        # If return, return after cachefile generated (used for tests)
        if use_cached == 'return':
            return

        df = load_h5_cache_file(cachefile, use_cached, metadata=metadata,
                             current_version=radis.__version__,
                             last_compatible_version=OLDEST_COMPATIBLE_VERSION,
                             verbose=verbose)

        if df is None:   # Read normal file
            df = pd.read_csv(energy_levels, comment='#', delim_whitespace=True)
            df = self._add_degeneracies(df)
            df = self._add_levels(df)

        self.df = df  # Store

        if use_cached and not exists(cachefile):
            save_to_hdf(self.df, cachefile, metadata=metadata, version=radis.__version__,
                        key='df', overwrite=True, verbose=verbose)

        # Placeholder: correct energies for isotope 2
        # Note that this is done on loaded database, never on the saved file
        if isotope == 2:
            self.df.grot *= 2
            if verbose:
                print('Warning. Energies for isotope 2 are assumed to be equal '
                      + 'to isotope 1 (degeneracy gs is still 2 instead of 1)')
                # TODO. ask Tashkun for energy levels of all isotopes

    def _add_degeneracies(self, df):
        ''' Calculate and store degeneracies in database df

        Parameters
        ----------

        df: pandas Dataframe
            energy database

        Notes
        -----

        .. warning::

            we use the same energies as CO2 626 but with gi = 2 for CO2 isotope 2 (636). 
            It is done in the __init__ method
        '''
        # Rotational degeneracy
        gj = 2*df.j+1
        # ... state dependant (forbidden rotational level are not in the database):
        gs = 1
        # ... state independant:
        gi = 1             
        grot = gj*gs*gi

        # Vibrational degeneracy
        gvib = 1           # CDSD is rovibrational complete

        # Store
        df['gvib'] = gvib
        df['gj'] = gj
        df['grot'] = grot

        return df

    def _add_levels(self, df):

        viblvl_label = self.viblvl_label

        if viblvl_label == 'p':
            df['viblvl'] = vib_lvl_name_cdsd_p(df.p, )
        elif viblvl_label == 'pc':
            df['viblvl'] = vib_lvl_name_cdsd_pc(df.p, df.c)
        elif viblvl_label == 'pcN':
            df['viblvl'] = vib_lvl_name_cdsd_pcN(df.p, df.c, df.N)
        elif viblvl_label == 'pcJN':
            df['viblvl'] = vib_lvl_name_cdsd_pcJN(df.p, df.c, df.j, df.N)
        elif viblvl_label is None:
            # dont label the levels. Wont be able to use the EnergyDatabase to fetch
            # vibrational energies for lines, however it can still be used to 
            # calculate Partition functions independently from a Spectrum calculation
            pass 
        else:
            raise ValueError(
                'Unexpected viblvl_label value: {0}'.format(viblvl_label))

        return df

#    def gvib(self):
#        ''' Vibrational degeneracy
#        1 in CDSD... but if using a HITRAN (v1v2'l2'v3) denomination remember
#        that gvib = v2+1 for CO2
#        '''
#
#        return 1

    def gs(self):
        from radis.db.degeneracies import gs
        I = self.isotope
        return gs(2, I)

    def gi(self):
        from radis.db.degeneracies import gi
        I = self.isotope
        return gi(2, I)


# %% Test
if __name__ == '__main__':

    from radis.test.levels.test_partfunc import _run_testcases
    print('Testing parfunc: {0}'.format(_run_testcases()))



# -*- coding: utf-8 -*-
"""
Created on Tue Jul 18 16:15:03 2017

@author: erwan

Summary
-------

Define :py:class:`~radis.db.classes.Molecule`, :py:class:`~radis.db.classes.Isotope` 
and :py:class:`~radis.db.classes.ElectronicState` classes 

ElectronicState has:

- Molecular parameters (read from HITRAN molparams.txt)
- Vibrational and rotational energies

Calculate Vibrational and Rotational energies for some database molecules

Molecules
-----------

Shortcut to some molecules ground states are defined at the end of the file 
in the dictionary `Molecules`, with the following structure: 

{molecule_name: {isotope_number: {electronic_state_name: ElectronicState object}}}
        
Import with::
    
    from radis.db.molecules import Molecules
    CO2_X = Molecules['CO2'][1]['X']            # 1 for first isotope

Or::

    from radis import getMolecule                # directly
    CO2_2 = getMolecule('CO2', 1, 'X')         # better KeyError messages 


-------------------------------------------------------------------------------


"""

from __future__ import division, absolute_import, print_function, unicode_literals
import re
from six import string_types
from radis.io.hitran import get_molecule, get_molecule_identifier
from radis.io.hitran import HITRAN_CLASS1, HITRAN_CLASS5
from radis.levels.dunham import Gv, Fv, EvJ
from radis.db.conventions import get_convention

# %% Molecule class

class Molecule(object):
    ''' Define a new molecule

    Parameters
    ----------

    name: str, int
        molecule name, or HITRAN identifier

    Other Parameters
    ----------------
    
    verbose: boolean
        more blabla
            
    See Also
    --------
    
    :py:class:`~radis.db.classes.Isotope`, :py:class:`~radis.db.classes.ElectronicState`,
    :py:func:`~radis.db.molecules.getMolecule`

    '''
    
    def __init__(self, name, verbose=True):    # @dev: no optional kwargs here (final stop)

        self.verbose = verbose

        # Get name and integer id
        if isinstance(name, string_types):
            self.name = name

            # Get name without parenthesis (without state) for HITRAN identification
            filtername = re.sub("[\(\[].*?[\)\]]", "", name)

            try:
                self.id = get_molecule_identifier(filtername)
            except KeyError:  # Not an HITRAN molecule
                self.id = None
        elif type(name) == int:
            self.id = name
            self.name = get_molecule(name)
        else:
            raise ValueError('Wrong name type:', name)


# %% Isotope class

class Isotope(Molecule):
    ''' Create an isotope of a given :py:class:`~radis.db.classes.Molecule`

    Parameters
    ----------

    molecule_name: str, or int
        molecule name or HITRAN identifier

    isotope: int
        isotope identifier, sorted by decreasing abundance (cf HITRAN 
        nomenclature)
        
    Other Parameters
    ----------------

    isotope_name: str
        (optional) isotope name. Default ''

    abundance: float
        isotopologue abundance. Default ``None``

    See Also
    --------
    
    :py:class:`~radis.db.classes.ElectronicState`, :py:class:`~radis.db.classes.Molecule`,
    :py:func:`~radis.db.molecules.getMolecule`
    '''

    def __init__(self, molecule_name, isotope, isotope_name='', abundance=None,
                 **kwargs):

        super(Isotope, self).__init__(
            name=molecule_name, **kwargs)  # initialize Molecule

        # Check input
        if not isinstance(isotope, int):
            raise TypeError('Wrong format for isotope {0}: expected int, got {1}'.format(
                isotope, type(isotope)))

        # Store
        self.iso = isotope
        self.isotope_name = isotope_name
        self.abundance = abundance

# %% ElectronicState class


class ElectronicState(Isotope):
    ''' Define an electronic state of an :py:class:`~radis.db.classes.Isotope` 
    of a :py:class:`~radis.db.classes.Molecule` 

    Parameters
    ----------

    molecule_name: str, int
        molecule name, or HITRAN identifier

    isotope: int
        isotope number of the molecule (sorted by abundance on Earth, HITRAN
        convention).

    state: str
        Electronic state name

    term_symbol: str
        Term symbol. Default ''

    g_e: int
        Degeneracy. Default ``None``

    spectroscopic_constants: dict
        spectroscopic constants for given (molecule, isotope, electronic state)
        Expected:
            
            Yij: cm-1
                rovibrational coefficients in Dunham convention
            
        or
        
            wexe, Be, etc. : cm-1
                rovibrational coefficients in Herzberg convention
            
        or 
                  
            Te: cm-1
                electronic energy. Default ``None`` if not given
                
        The default RADIS :ref:`spectroscopic constants <label_db_spectroscopic_constants>` 
        can be used, but others can be supplied too.

    Erovib: function
        Rovibrational energy of the molecule. Accessible with, typically, Mol.Erovib(v,J)

    Ehaj: function
        To return vibrational energy in harmonic and anharmonic component, 
        plus rotational energy. Used to build Treanor distributions

    vmax, vmax_morse: int, or ``None``
        maximum vibrational number (required for partition function calculation)
        to be calculated with Dunham expansion, and Morse potential
        If None, number will be infered from dissociation energy

    Jmax: int, or ``None``
        maximum rotational number (required for partition function calculation)
        If None, number will be infered from dissociation energy

    Ediss: cm-1
        dissociation energy. Required for partition function calculation 
        if neither vmax nor Jmax are given

    kwargs: **dict
        forwarded to parent class

    See Also
    --------
    
    :py:class:`~radis.db.classes.Isotope`, :py:class:`~radis.db.classes.Molecule`,
    :py:func:`~radis.db.molecules.getMolecule`

    '''

    #''' 
    # TODO
    #To allow for more flexibility, ElectronicState should also inherit from some 
    #built-in Symmetry / species class, Sigma / Pi / Delta etc... where the energy 
    #is calculated appropriately.  (ex: Rigid Rotor, Symmetric Top etc.)

    # (major code rewritting ahead...)
    # That should be part of RADIS extension to non Sigma electronic states. 


    def __init__(self, molecule_name, isotope, state, term_symbol='',
                 g_e=None, spectroscopic_constants={}, Erovib=None, Ehaj=None,
                 vmax=None, vmax_morse=None, Jmax=None, Ediss=None, **kwargs):

        super(ElectronicState, self).__init__(molecule_name=molecule_name,
                                              isotope=isotope, **kwargs)    # initialize Isotope

        # Initialize electronic state specific parameters
        self.state = state
        self.term_symbol = term_symbol
        self.g_e = g_e

        self._parse_rovib_constants(spectroscopic_constants)

#        self.Erovib = None   #: func: overwritten by Erovib if given, or by default Dunham developments if spectroscopic_constants are given
        self.Ehaj = None     #: func: overwritten by Erovib if given, or by default Dunham developments if spectroscopic_constants are given
        self._assign_E(Erovib, Ehaj)
        self.state = state
        self.vmax = vmax
        self.vmax_morse = vmax_morse
        self.Jmax = Jmax
        self.Ediss = Ediss

    def _parse_rovib_constants(self, spectroscopic_constants):
        ''' Parse spectroscopic constants 
        
        Stores :py:attr:`~radis.db.classes.ElectronicState.Te` and 
        :py:attr:`~radis.db.classes.ElectronicState.re` as electronic state 
        attributes, and the rest under :py:attr:`~radis.db.classes.ElectronicState.rovib_constants` 
        
        Parameters
        ----------
        
        spectroscopic_constants: dict
            spectro constants in Herzberg or Dunham convention
            
        Returns
        -------
        
        None: 
            but constants are stored under :py:attr:`~radis.db.classes.ElectronicState.rovib_constants`
        
        '''
        
        import re
        # Clean keys
        # In particular, remove trailing '_cm-1' if given in dict or database 
        rovib_constants = {re.sub('_cm-1$', '', k):v 
                           for (k,v) in spectroscopic_constants.items()}
        
        # Get specific keys
        self.Te = rovib_constants.pop('Te', None)  # default None
        self.re = rovib_constants.pop('re', None)
        
        # Store
        self.rovib_constants = rovib_constants

    # Default method to calculate energy

    def _assign_E(self, Erovib, Ehaj):
        ''' Finds appropriate Electrorovibrational energy function for this molecule ,
        based on which convention is used for spectroscopic coefficients 
        (Dunham, or Herzberg)
        
        Replaces the :py:meth:`~radis.db.classes.ElectronicState.Erovib` template
        with the appropriate function for this molecule / isotope / electronic state.
        
        Default methods only work for diatomic molecules. For polyatomic molecules,
        you should give the energy function directly (see how CO2 is defined in 
        radis.db.molecules.py) 
        
        Returns
        -------
        
        None:
            but assigns the ``self.Erovib`` function to the molecule
            
        '''
        # Check input
        if Erovib is None and Ehaj is not None:
            raise ValueError("If giving Ehaj (harmonic and anharmonic component calculation) "+\
                             "you must also supply Erovib (total rovibrational energy)")
        
        if Erovib is not None:
            # overwrite energy calculation
            self.Erovib = Erovib
            self.Ehaj = Ehaj
            if self.verbose>=2:
                print('{0}: overwritting Energy calculation with {1}'.format(
                        self.get_fullname(), Erovib))
        else:
            # Autofind which energy model to use for the given molecule
            c = self.rovib_constants
            if len(c) == 0:
                pass    # keep the default function, that will raise an error on first call.
                if self.verbose>=2:
                    print('{0}: No rovibconstants found'.format(
                            self.get_fullname()))
            else: 
                convention = get_convention(c)   # Herzberg or Dunham
                    
                if self.name in HITRAN_CLASS1:
                    if convention == 'herzberg':
                        self.Erovib = self._E_Herzberg
                        self.Ehaj = None     # NotImplemented
                        if self.verbose>=2:
                            print('{0}: using Herzberg coefficients for Energy calculation'.format(
                                    self.get_fullname()))
                    else:
                        self.Erovib = self._E_Dunham
                        self.Ehaj = None     # NotImplemented
                        if self.verbose>=2:
                            print('{0}: using Dunham coefficients for Energy calculation'.format(
                                    self.get_fullname()))
                        # TODO: reformat these methods as external functions, but keep docstrings
                elif self.name in HITRAN_CLASS5:
                    if convention == 'herzberg':
                        from radis.levels.energies_co2 import EvJ_co2, EvJah_co2
                        self._Erovib = EvJ_co2
                        self._Ehaj = EvJah_co2
                        self.Erovib = self._Erovib_default_coefs
#                        self.Erovib.__doc__ == EvJ_co2.__doc__
                        self.Ehaj = self._Ehaj_default_coefs
#                        self.Ehaj.__doc__ == EvJah_co2.__doc__
                    else:
                        raise NotImplementedError('Only Herzberg convention spectroscopic '+\
                                                  'constants defined for {0}. '.format(self.name)+\
                                                  'You still define your own Erovib function '+\
                                                  'and overwrite the ElectronicState class one.')
                        
    def Erovib(self, v, J, remove_ZPE=True):
        ''' Calculate rovibrational energy of molecule 
        
        .. math::
            
            E(v,J) = G(v) + F(v,J)
            
        Works with Herzberg convention for the spectroscopic coefficients used 
        in :func:`~radis.levels.dunham.Gv`, :func:`~radis.levels.dunham.Fv` ::
        
        Parameters
        ----------

        v: int
            vibrational state 
            For polyatomic molecules, use ``v1, v2, v3, etc.`` instead of ``v``. 
        
        J: int
            rotational state

        remove_ZPE: boolean
            if ``True``, removes energy of v=0,J=0 vibrational level (zero-point-energy)
            Default ``True``

        Returns
        -------

        energy of state in cm-1 
        
        Notes
        -----
        
        Method is overwritten on molecule creation for polyatomic molecules. 
        Refer to the correct method below: 
            
        - CO2: :py:func:`~radis.levels.energies_co2.EvJ_co2`
        
        
        See Also
        --------

        :func:`~radis.levels.dunham.Gv`, :func:`~radis.levels.dunham.Fv`

        
        '''
        
        raise NotImplementedError('Rovibrational energy not implemented for {0}'.format(
                self.get_fullname))  # @dev: see radis.db.classes.ElectronicState._assign_E
        
    def _Erovib_default_coefs(self, *args, **kwargs):
        ''' Calculate rovibrational energy of molecule 
        
        .. math::
            
            E(v,J) = G(v) + F(v,J)
            
        Works with Herzberg convention for the spectroscopic coefficients used 
        in :func:`~radis.levels.dunham.Gv`, :func:`~radis.levels.dunham.Fv` ::
        
        Parameters
        ----------

        v: int
            vibrational state 
            For polyatomic molecules, use ``v1, v2, v3, etc.`` instead of ``v``. 
        
        J: int
            rotational state

        remove_ZPE: boolean
            if ``True``, removes energy of v=0,J=0 vibrational level (zero-point-energy)
            Default ``True``

        Returns
        -------

        energy of state in cm-1 
        
        Notes
        -----
        
        Method is overwritten on molecule creation for polyatomic molecules. 
        Refer to the correct method below: 
            
        - CO2: :py:func:`~radis.levels.energies_co2.EvJ_co2`
        
        
        See Also
        --------

        :func:`~radis.levels.dunham.Gv`, :func:`~radis.levels.dunham.Fv`

        '''
        assert 'coeff_dict' not in kwargs
        kwargs.update({'coeff_dict':self.rovib_constants})
        
        return self._Erovib(*args, **kwargs)
        
    def _Ehaj_default_coefs(self, *args, **kwargs):
        ''' Call default _Ehaj() function with default rovib_constants coefficients
        '''
        
        assert 'coeff_dict' not in kwargs
        kwargs.update({'coeff_dict':self.rovib_constants})
        
        return self._Ehaj(*args, **kwargs)
        
    def _E_Dunham(self, v, J, remove_ZPE=True):
        # TODO: move in levels.energies as was done for CO2
        
        c = self.rovib_constants
        
        if remove_ZPE:
            ZPE = EvJ(0, 0, **c)
        else:
            ZPE = 0

        return EvJ(v, J, **c) - ZPE
        
    def _E_Herzberg(self, v, J, remove_ZPE=True):
        # TODO: move in levels.energies as was done for CO2
        ''' Calculate rovibrational energy of molecule 
        
        .. math::
            
            E(v,J) = G(v) + F(v,J)
            
        Works with Herzberg convention for the spectroscopic coefficients used 
        in :func:`~radis.levels.dunham.Gv`, :func:`~radis.levels.dunham.Fv` ::
        
        Only works for diatomic molecules. Method should be overwritten on molecule
        creation for other molecules 

        Parameters
        ----------

        v: int
            vibrational state 

        J: int
            rotational state

        remove_ZPE: boolean
            if ``True``, removes energy of v=0,J=0 vibrational level (zero-point-energy)
            Default ``True``

        Returns
        -------

        energy of state in cm-1 
        
        See Also
        --------

        :func:`~radis.levels.dunham.Gv`, :func:`~radis.levels.dunham.Fv`
        
        '''
        

        try:

            # Get from constants (hardcoded, see reference above)
            c = self.rovib_constants
            
            # Vibrational constants
            we = c['we']                 # mandatory
            wexe = c.get('wexe', 0)      # optional: 0 is default value
            weye = c.get('weye', 0)
            weze = c.get('weze', 0)
            weae = c.get('weae', 0)
            webe = c.get('webe', 0)

            # Rotational constants
            Be = c['Be']
            De = c.get('De', 0)

            alpha_e = c.get('alpha_e', 0)
            beta_e = c.get('beta_e', 0)
            gamma_e = c.get('gamma_e', 0)
            delta_e = c.get('delta_e', 0)
            pi_e = c.get('pi_e', 0)

            He = c.get('He', 0)
            eta_e = c.get('eta_e', 0)

            # Energies
#            Te = c['Te']      # electronic energy
            G = Gv(v, we, wexe, weye, weze, weae, webe)   # vibrational energy
            F = Fv(v, J, Be, De, alpha_e, beta_e, gamma_e, delta_e, pi_e=pi_e,
                   He=He, eta_e=eta_e)    # rotational energy

            if remove_ZPE:
                ZPE = Gv(0, we, wexe, weye, weze)
            else:
                ZPE = 0

            E = G + F - ZPE    # cm-1

        except KeyError as err:
            raise KeyError('Mandatory spectroscopic constant `{0}` '.format(err.args[0]) +
                           'not defined for electronic state {0}'.format(self.get_fullname()) +
                           '. Check your ElectronicState definition')

        return E

    # %% Informative methods

    def get_Morse_inc(self):
        ''' Get Morse potential energy increment correction for given molecule

        Examples
        --------
        
        ::

            inc = ElecState.get_Morse_inc()
            Delta_E(vi, vi+1) = Delta_E0 - (vi+1-vi0)*inc

        for the energy gap between vibrational level vi and vi+1, 
        where vi0 is the last vibrational level calculated with Dunham 
        expansion, and  and Delta_E0 = DeltaE(vi0-1, vi0)
        
        See Also
        --------
        
        :py:func:`~radis.phys.morse.morse_increment`

        '''
        
        from radis.phys.morse import morse_increment
        
        c = self.rovib_constants

        if get_convention(c) == 'dunham':
        
            # convert to get wexe, weae etc in Herzberg convention
            from radis.db.conventions import herzberg2dunham
            def convert(coef, default=0):
                if coef in c:
                    sign, Yij = herzberg2dunham[coef]
                    return sign*Yij
                else:
                    return default    # default
            
            we = convert('we')             # required
            wexe = convert('wexe', 0)  # optional: 0 is default value
            weye = convert('weye', 0)
            weze = convert('weze', 0)
            weae = convert('weae', 0)
            webe = convert('webe', 0)
            wece = convert('wece', 0)
            
            if we is None:
                raise KeyError('Mandatory spectroscopic constant `we` ' +
                               'not defined for electronic state {0}'.format(self.get_fullname()) +
                               '. Check your ElectronicState definition. Given '+\
                               'constants: {0}'.format(list(c.keys())))

        else:
            try:
                we = c['we']             # required
                wexe = c.get('wexe', 0)  # optional: 0 is default value
                weye = c.get('weye', 0)
                weze = c.get('weze', 0)
                weae = c.get('weae', 0)
                webe = c.get('webe', 0)
                wece = c.get('wece', 0)
                    
            except KeyError as err:
                raise KeyError('Mandatory spectroscopic constant `{0}` '.format(err.args[0]) +
                               'not defined for electronic state {0}'.format(self.get_fullname()) +
                               '. Check your ElectronicState definition. Given '+\
                               'constants: {0}'.format(list(c.keys())))

        return morse_increment(self.Ediss, we, wexe=wexe, weye=weye, weze=weze,
                               weae=weae, webe=webe, wece=wece)

    def get_fullname(self):
        return '{0}({1}{2})(iso{3})'.format(self.name, self.state,
                                            self.term_symbol, self.iso)

    def get_statename_utf(self):
        ''' Returns Electronic state name in UTF format, used in the standard
        RADIS JSON databases. Examples::
            
           'X', '1Σ+' 
           >>> X1SIG+
           
           'C', '3Πu' 
           >>> C3PIu
           
        '''
        
        state = self.state
        term = self.term_symbol
        
        replace_utf = {'Σ':'SIG',
                       'Π':'PI',
                       'Δ':'DEL'}
        
        for k, v in replace_utf.items():
            term = term.replace(k, v)
            
        return state+term
    
    
    

# %% Test


if __name__ == '__main__':
    
    from radis.test.db.test_molecules import _run_testcases
    print(('Testing molecules.py', _run_testcases()))



# -*- coding: utf-8 -*-
"""
Created on Mon Nov 20 09:38:23 2017

@author: erwan

Use
------

Run all tests:

>>> pytest       (in command line, in project folder)

Run only fast tests (i.e: tests that a 'fast' label)

>>> pytest -m fast

"""

from __future__ import print_function, absolute_import, division, unicode_literals
from radis.lbl.parallel import ParallelFactory, SpectrumFactory
from radis.misc.printer import printm, printr
from radis.misc.utils import DatabankNotFound
from radis.test.utils import IgnoreMissingDatabase, setup_test_line_databases
from time import time
from warnings import warn
from six.moves import range
import pytest

# %% Main


@pytest.mark.python3_only
def test_parallel(plot=False, verbose=True, warnings=True, *args, **kwargs):
    ''' Core of the parallel test routine 
    
    Test than parallel is at least faster than sequential mode. Note that 
    this is not necessary: parallelization becomes much faster during the 
    convolution steps, hence for large number of lines
    '''

    if plot:  # Make sure matplotlib is interactive so that test are not stuck in pytest
        import matplotlib.pyplot as plt
        plt.ion()
        
    setup_test_line_databases()

    try:
        # Performance test
        Tvib = 1000  # , 1200, 1400]
        Trot = [900, 1200, 1250, 1300, 1400, 1600]
        path_length = 0.1
        mole_fraction = 0.5
        
        cutoff = 0
        broadening_max_width = 50
        wstep = 0.01

        # .. parallel
        pf = ParallelFactory(wavelength_min=4165, wavelength_max=4200,
                             cutoff=cutoff,
                             db_use_cached=True,
                             lvl_use_cached=True,
                             isotope='1',
                             broadening_max_width=broadening_max_width,
                             wstep=wstep,
                             verbose=verbose)
        pf.warnings['MissingSelfBroadeningWarning'] = 'ignore'
        pf.warnings['NegativeEnergiesWarning'] = 'ignore'
        pf.warnings['EmptyDatabaseWarning'] = 'ignore'
        pf.load_databank('HITRAN-CO2-TEST')

        t0 = time()
        s_parallel = pf.non_eq_spectrum(Tvib, Trot, path_length=path_length, 
                                        mole_fraction=mole_fraction)
        t0 = time() - t0

        # ... non parallel
        sf = SpectrumFactory(wavelength_min=4165, wavelength_max=4200,
                             cutoff=cutoff,
                             db_use_cached=True,
                             lvl_use_cached=True,
                             isotope='1,2',
                             broadening_max_width=broadening_max_width,
                             wstep=wstep,
                             verbose=verbose)
        sf.warnings['MissingSelfBroadeningWarning'] = 'ignore'
        sf.warnings['NegativeEnergiesWarning'] = 'ignore'
        sf.warnings['EmptyDatabaseWarning'] = 'ignore'
        sf.load_databank('HITRAN-CO2-TEST')
        t1 = time()
        s_sequential = []
        for Troti in Trot:
            s_sequential.append(sf.non_eq_spectrum(Tvib, Troti, path_length=path_length,
                                                   mole_fraction=mole_fraction))
        t1 = time() - t1

        if verbose:
            printm(
                ('\nParallelFactory performance test ({0} cases)'.format(len(Trot))))
            printm('-------------------')
            printm(('Parallel version: {0:.1f}s'.format(t0)))
            printm(('Normal version: {0:.1f}s'.format(t1)))
            printm('-------------------')

        # Assert that spectra are the same
        for i in range(len(Trot)):
            assert s_parallel[i].compare_with(s_sequential[i], spectra_only=True,
                             plot=False)
        if verbose:
            printm('Ensured that spectra are the same: OK')
            
#        # parallel version should be at least faster!
#        # note: ...  may not be the case if the calculated ranges are too small
#        # ......... as is the case in the test database. Better use // for spectra
#        # ......... with calculation times > 5s
#        assert t0 < t1
        if t0 < t1:
            warn('Parallel version ({0:.1f}s) is slower than the Serial version ({0:.1f}s) !'.format(
                    t0, t1))

        return True  # no plot implemented

    except DatabankNotFound as err:
        assert IgnoreMissingDatabase(err, __file__, warnings)


if __name__ == '__main__':

    printm(('Tested parallel.py:', test_parallel(plot=True)))



# -*- coding: utf-8 -*-
"""
Created on Thu May 28 14:47:36 2015

@author: Erwan
"""

from __future__ import print_function, absolute_import, division, unicode_literals

import os
import json
from collections import OrderedDict
from radis.misc.basics import is_number


def getFile(*relpath):
    ''' Converts the relative path of a database file in a the full path.
    Used by processing script not to worry about where the database is stored

    Examples of use::
        
        radis.db.getFile('CN','CN_Violet_vib.dat')
        radis.db.getFile('CN\CN_Violet_vib.dat')    
        radis.db.getFile('CN\\CN_Violet_vib.dat')

    '''
    
#    return os.path.join(os.path.dirname(__file__), *relpath)
    from radis.misc.utils import getProjectRoot
    return os.path.join(getProjectRoot(), 'db', *relpath)



def check_molecule_data_structure(fname, verbose=True):
    ''' Check that ``fname`` has a valid JSON structure for molecular data
    
    Parameters
    ----------
    
    fname: str
        molecular data JSON file 
        
        
    Notes
    -----
    
    Order in the json doesnt matter, however, the order of the 
    electronic_level_names matters: the index of all levels should
    match the `index` key of these states. 
    
    '''

    with open(fname) as f:
        db = json.load(f) #, object_pairs_hook=OrderedDict)
            
    for molecule, db_molec in db.items():
        # ... Check number of isotopes is correct
        isotope_names = db_molec['isotopes_names']
        isotopes = db_molec['isotopes']
        if len(isotopes) != len(isotope_names):
            raise ValueError(
                    'In molecule {0}: isotope names '.format(molecule)+\
                    '({0}) dont match the number of isotopes ({1})'.format(
                    isotope_names, list(isotopes.keys())))
            
        # ... Check number of electronic states is correct
        for isotope, db_iso in db_molec['isotopes'].items():
            elec_states_names = db_iso['electronic_levels_names']
            elec_states = db_iso['electronic_level']
            if len(elec_states_names) != len(elec_states):
                raise ValueError(
                    'In molecule {0}, isotope {1}: electronic '.format(molecule, isotope)+\
                    'levels names ({0}) dont match the number of levels ({1})'.format(
                    elec_states_names, list(elec_states.keys())))
        
        # ... Check they are properly ordered
            for state, db_state in elec_states.items():
                if elec_states_names.index(db_state['name'])+1 != db_state['index']:  
                    # ... + 1 because Python index start at 0 and FORTRAN at 1 unless allocated with 0:N
                    raise ValueError(
                        'In molecule {0}, isotope {1}: index of electronic '.format(
                                molecule, isotope, db_state['index'])+\
                        'state {0} ({1}): {2} does not match the list of states: {3}. '.format(
                        state, db_state['name'], db_state['index'], elec_states_names))
        
    if verbose: print('Structure of {0} looks correct'.format(fname))


def get_rovib_coefficients(molecule, isotope, electronic_state, jsonfile='default',
                           remove_trailing_cm1=True):
    ''' Returns all rovib coefficients for ``molecule``, ``isotope``, ``electronic_state`` 
    by parsing a JSON file of molecule data. 
    
    Parameters
    ----------
    
    molecule: str
        molecule name
        
    isotope: int
        isotope number
        
    electronic_state: str
        electronic state name
        
    jsonfile: str, or ``default``
        path to json file. If ``default``, the ``molecules_data`` JSON file 
        in the ``radis.db`` database is used::
        
            radis\db\[molecule]\molecules_data.json
            
    Other Parameters
    ----------------
    
    remove_trailing_cm1: bool
        if ``True``, remove the trailing '_cm-1' in the spectroscopic coefficient name, 
        if defined. Ex:: 
            
            ('we1_cm-1', 1333.93)
            >>> ('we1', 1333.93)
    
    See Also
    --------
    
    :py:func:`~radis.db.utils.get_dunham_coefficients`, 
    :py:func:`~radis.db.utils.get_herzberg_coefficients`, 
    '''

    if jsonfile == 'default':
        jsonfile = getFile('{0}/molecules_data.json'.format(molecule))
        
    check_molecule_data_structure(jsonfile, verbose=False)
    
    with open(jsonfile) as f:
        db = json.load(f, object_pairs_hook=OrderedDict)

    # Get Dunham coefficients in 001 state (X)
    elec_state_names = db[molecule]['isotopes'][str(isotope)]['electronic_levels_names']
    
    try:
        elec_state_index = elec_state_names.index(electronic_state)
    except ValueError:
        raise ValueError("{0} not in the electronic state list for {1}(iso={2}): {3}".format(
                electronic_state, molecule, isotope, elec_state_names))
    else:
        elec_state_index = '{:03d}'.format(elec_state_index+1)  # 1 based index
    
    rovib_coeffs = db[molecule]['isotopes'][str(isotope)]['electronic_level'][elec_state_index]
    
    if remove_trailing_cm1:
        
        def remove_cm1(coef):
            ''' Remove the trailing '_cm-1' in the spectroscopic coefficient name, 
            if defined '''
            if coef.endswith('_cm-1'):
                coef = coef[:-5]
            return coef
                
        rovib_coeffs = {remove_cm1(k):v for (k,v) in rovib_coeffs.items()}
        
    return rovib_coeffs

def get_dunham_coefficients(molecule, isotope, electronic_state, jsonfile='default'):
    ''' Returns Dunham coefficients ``Yij`` for ``molecule``, ``isotope``, ``electronic_state`` 
    by parsing a JSON file of molecule data. 
    
    Dunham coefficients are identified as starting with ``Y`` (i.e. ``alpha_e``
    won't be recognized)
    
    Parameters
    ----------
    
    molecule: str
        molecule name
        
    isotope: int
        isotope number
        
    electronic_state: str
        electronic state name
        
    jsonfile: str, or ``default``
        path to json file. If ``default``, the ``molecules_data`` JSON file 
        in the ``radis.db`` database is used::
        
            radis\db\[molecule]\molecules_data.json
    
    '''

    rovib_coeffs = get_rovib_coefficients(molecule, isotope, electronic_state, jsonfile=jsonfile)
    
    # Only get Dunham coeffs, i.e, these that start with Y
    dunham_coeffs = {k:v for (k, v) in rovib_coeffs.items() if k.startswith('Y')}

    if len(dunham_coeffs) == 0:
        raise ValueError('No Dunham coefficients found for {0}{1}(iso={2})'.format(
                molecule, isotope, electronic_state))

    return dunham_coeffs

def get_herzberg_coefficients(molecule, isotope, electronic_state, jsonfile='default'):
    ''' Returns spectroscopic coefficients with Herzberg conventions for 
    ``molecule``, ``isotope``, ``electronic_state``  by parsing a JSON file of molecule data. 
    
    Herzberg coefficients are the usual:
        
    .. math::
        
        \\omega_e, \\alpha_e, B_e, D_e, etc.
    
    The extensive list of parameters considered as Herzberg coefficients is found in 
    :py:data:`~radis.db.conventions.herzberg_coefficients`
    
    Parameters
    ----------
    
    molecule: str
        molecule name
        
    isotope: int
        isotope number
        
    electronic_state: str
        electronic state name
        
    jsonfile: str, or ``default``
        path to json file. If ``default``, the ``molecules_data`` JSON file 
        in the ``radis.db`` database is used::
        
            radis\db\[molecule]\molecules_data.json
    
    See Also
    --------
    
    :py:data:`~radis.db.conventions.herzberg_coefficients`
    
    '''
    
    from radis.db.conventions import herzberg_coefficients

    rovib_coeffs = get_rovib_coefficients(molecule, isotope, electronic_state, jsonfile=jsonfile)
    
#    # Only get Herzberg coeffs, i.e, these that are defined in :py:data:`~radis.db.conventions.herzberg_coefficients`
    
    herzberg_coeffs = {k:v for (k, v) in rovib_coeffs.items() if ignore_trailing_number(k) in herzberg_coefficients}

    if len(herzberg_coeffs) == 0:
        raise ValueError('No Herzberg spectroscopic coefficients found for {0}{1}(iso={2})'.format(
                molecule, isotope, electronic_state))

    return herzberg_coeffs

def ignore_trailing_number(coef):
    ''' Used so that ``wexe1`` matches ``wexe`` as a well defined 
    Herzberg coefficient '''
    
    if is_number(coef[-1]):
        coef = coef[:-1]
    return coef



# -*- coding: utf-8 -*-
"""
Created on Mon Nov 20 09:59:55 2017

@author: erwan

Examples
--------

Run all tests::

    pytest       (in command line, in project folder)

Run only fast tests (i.e: tests that have a 'fast' label)::

    pytest -m fast
    
------------------------------------------------------------------------

"""

from __future__ import unicode_literals, print_function, absolute_import, division
from radis.lbl import SpectrumFactory
from radis.misc.printer import printm
from radis.misc.utils import DatabankNotFound
from radis.test.utils import IgnoreMissingDatabase, setup_test_line_databases
from radis.spectrum import Spectrum
import numpy as np
import matplotlib.pyplot as plt
from os.path import basename
import pytest

fig_prefix = basename(__file__)+': '

# %% ======================================================================
# TEST
# ---------------------


@pytest.mark.needs_config_file
@pytest.mark.needs_db_CDSD_HITEMP
def test_spec_generation(plot=True, verbose=2, warnings=True, *args, **kwargs):
    ''' Test spectrum generation
    Can be used as a base to generate spectra in your codes

    Non-regression test: compare with past version (see conditions below)

    Compare results from a reference case to results calculated on 30/12/2017
    This is not a validation case (30/12/2017 results are not a physically validated
    case), but it makes sure results dont change over time

    Conditions (30/12/2017)::

        Physical Conditions
        ----------------------------------------
           Tgas                 300 K
           Trot                 300 K
           Tvib                 300 K
           pressure             1.01325 bar
           isotope              1,2
           mole_fraction        1
           molecule             CO2
           path_length          1 cm
           wavelength_max       4400.0 nm
           wavelength_min       4150.0 nm
           wavenum_max          2409.6385542168673 cm-1
           wavenum_min          2272.7272727272725 cm-1
        Computation Parameters
        ----------------------------------------
           Tref                 296 K
           broadening_max_width  10 cm-1
           cutoff               1e-25 cm-1/(#.cm-2)
           db_assumed_sorted    True
           db_use_cached        True
           dbformat             cdsd
           dbpath               # USER-DEPENDANT: CDSD-HITEMP
           fillmissinglevelswithzero  False
           levelsfmt            cdsd
           levelspath           # USER-DEPENDANT: CDSD-4000
           medium               vacuum
           parfuncfmt           cdsd
           parfuncpath          # USER-DEPENDANT: CDSD-4000
           rot_distribution     boltzmann
           self_absorption      True
           vib_distribution     boltzmann
           wavenum_max_calc     2414.6385542168673 cm-1
           wavenum_min_calc     2267.7272727272725 cm-1
           waveunit             cm-1
           wstep                0.01 cm-1
        ----------------------------------------

    Notes
    -----

    Performance test. How long it tooks to calculate this Spectrum?
    Test with cutoff 1e-25, broadening_max_width=10 

    - 0.9.15: >>> 33s

    - 0.9.16*: (replaced groupby().apply() with iteration over indexes) >>> 32s  
            [but large impact expected on big files]

    - 0.9.16*: (upgraded cache files to h5) >>> 25s 

    - 0.9.16*: (also added h5 cache file for levels) >>> 21s

    - 0.9.16*: (with Whiting slit voigt function) >>> 5.8s

    Test with cutoff 1e-27, broadening_max_width=50 :  
    ("Spectrum calculated in ... ", including database loading time)

    - 0.9.16*: (same code as last) >>> 12.5s including 7.6s of broadening

    - 0.9.16**: (with pseudo_continuum_threshold=0.01) >>> 7.8s including 2.3s of broadening

    - 0.9.18 (normal code, no pseudo continuum). >>> ?
    
    - 0.9.21 (normal code) >>> 13.7s, including 8.7s of broadening
             (with pseudo_continuum_threshold=0.01) >>> 4.3s, including 2.6s of broadening

    - 0.9.21*              >>> 14.0s  (added the manual lineshape normalization instead of
                                       Whitings's polynomial)
    
    - 0.9.22 (normal code) >>> 11.3s   (without energy level lookup, for eq. calculations)
             (with pseudo_continuum_threshold=0.01) >>> 5.9s

    - 0.9.23 (normal code) >>> 7.2s   (added jit in Voigt broadening)
                           >>> 7.1s   (chunksize = None)  (and much faster for more lines)
             (with pseudo_continuum_threshold=0.01) >>> 4.9s
    
    '''

    if plot:  # Make sure matplotlib is interactive so that test are not stuck in pytest
        plt.ion()

    try:
        from time import time
        t0 = time()
        if verbose:
            printm('>>> _test_spec_generation')

         # This is how you get a spectrum (see calc.py for front-end functions
        # that do just that)
        sf = SpectrumFactory(wavelength_min=4150, wavelength_max=4400,
                             parallel=False, bplot=False, cutoff=1e-27,
                             isotope='1,2', db_use_cached=True,
                             broadening_max_width=50,
#                             pseudo_continuum_threshold=0.01,
                             medium='vacuum',
                             verbose=verbose)
        sf.warnings['MissingSelfBroadeningWarning'] = 'ignore'
        sf.warnings['NegativeEnergiesWarning'] = 'ignore'
        sf.load_databank('HITEMP-CO2-DUNHAM',
                         load_energies=False,  # no need to load energies at equilibrium
                         )
        s = sf.eq_spectrum(Tgas=300)
        if verbose:
            printm('>>> _test_spec_generation: Spectrum calculated in {0:.1f}s'.format(time()-t0))
            
        if plot:
            plt.figure(fig_prefix+'Reference spectrum CDSD-HITEMP (radiance)')
            # Iunit is arbitrary. Use whatever makes sense
            s.plot('radiance_noslit', Iunit='µW/cm2/sr/nm', nfig='same')
        s.rescale_path_length(0.01)

        # Here we get some extra informations:
        if plot:
            sf.plot_broadening(i=0)          # show broadening of one line
            plt.xlim((2267.20, 2268.30))

        # Compare with harcoded results
        # ... code previously used to export hardcoded results:
        # ... >>> np.savetxt('output.txt', np.vstack(s.get('abscoeff', wunit='nm', medium='air')).T[::10])
        # ... and header contains all input conditions:
        # ... >>> print(s)
        from radis.test.utils import getTestFile
        wref, Iref = np.loadtxt(getTestFile(
            'CO2abscoeff_300K_4150_4400nm.txt')).T
        match_reference = np.allclose(
            s.get('abscoeff', wunit='nm', medium='air')[1][::10], Iref)
        if not match_reference:
            # give some more information before raising error
            printm('Error: {0:.2f}%'.format(np.mean(abs(s.get('abscoeff', wunit='nm',
                                                              medium='air')[1][::10]/Iref-1))*100))
            # Store the faulty spectrum
            s.store('test_factory_failed_{0}.spec'.format(neq.get_version()),
                    if_exists_then='replace')

        # Plot comparison
        if plot:
            plt.figure(fig_prefix+'Reference spectrum (abscoeff)')
            # , show_points=True)  # show_points to have an
            s.plot('abscoeff', wunit='nm', medium='air',
                   nfig='same', lw=3, label='RADIS, this version')
            # idea of the resolution
            plt.plot(wref, Iref, 'or', ms=3,
                     label='version NEQ 0.9.20 (12/05/18)')
            plt.legend()
            plt.title('All close: {0}'.format(match_reference))
            plt.tight_layout()

        # Another example, at higher temperature.
        # Removed because no test is associated with it and it takes time for
        # nothing
#        s2 = sf.non_eq_spectrum(Tvib=1000, Trot=300)
#        if plot: s2.plot('abscoeff', wunit='nm')

        if verbose:
            printm('Spectrum calculation (no database loading) took {0:.1f}s\n'.format(
                s.conditions['calculation_time']))
            printm(
                '_test_spec_generation finished in {0:.1f}s\n'.format(time()-t0))

        assert match_reference

    except DatabankNotFound as err:
        assert IgnoreMissingDatabase(err, __file__, warnings)

# Test power integral


@pytest.mark.fast
def test_power_integral(verbose=True, warnings=True, *args, **kwargs):
    ''' Test direct calculation of power integral from Einstein coefficients
    matches integration of broadened spectrum in the optically thin case

    We compare:

    - direct calculation of power integral with equilibrium code
        :meth:`~radis.lbl.SpectrumFactory.optically_thin_power` (T)
    - direct calculation of power integral with non equilibrium code  
        :meth:`~radis.lbl.SpectrumFactory.optically_thin_power` (T,T)
    - numerical integration of non equilibrium spectrum under optically thin conditions: 
        :meth:`~radis.spectrum.spectrum.Spectrum.get_power`

    Test passes if error < 0.5%

    '''

    try:
        if verbose:
            printm('>>> _test_power_integral')

        setup_test_line_databases()  # add HITRAN-CO-TEST in ~/.radis if not there

        sf = SpectrumFactory(wavelength_min=4300, wavelength_max=4666,
                             wstep=0.001,
                             parallel=False, bplot=False, cutoff=1e-30,
                             path_length=10, mole_fraction=400e-6,
                             isotope=[1], db_use_cached=True,
                             broadening_max_width=10,
                             verbose=verbose)
        sf.warnings['MissingSelfBroadeningWarning'] = 'ignore'
        sf.load_databank('HITRAN-CO-TEST')
        unit = 'µW/sr/cm2'
        T = 600

        # Calculate:

        # ... direct calculation of power integral with equilibrium code
        Peq = sf.optically_thin_power(Tgas=T, unit=unit)

        # ... direct calculation of power integral with non equilibrium code
        Pneq = sf.optically_thin_power(Tvib=T, Trot=T, unit=unit)

        # ... numerical integration of non equilibrium spectrum under optically thin
        # ... conditions
        sf.input.self_absorption = False
        s = sf.non_eq_spectrum(T, T)

        assert s.conditions['self_absorption'] == False

        # Compare
        err = abs(Peq-s.get_power(unit=unit))/Peq
        if verbose:
            printm('Emission integral:\t{0:.4g}'.format(Peq), unit)
            printm('Emission (noneq code):\t{0:.4g}'.format(Pneq), unit)
            printm('Integrated spectrum:\t{0:.4g}'.format(
                s.get_power(unit=unit)), unit)
            printm('Error: {0:.2f}%'.format(err*100))

        assert err < 0.005

    except DatabankNotFound as err:
        assert IgnoreMissingDatabase(err, __file__, warnings)


@pytest.mark.fast
def test_media_line_shift(plot=False, verbose=True, warnings=True, *args, **kwargs):
    ''' See wavelength difference in air and vacuum '''

    if plot:  # Make sure matplotlib is interactive so that test are not stuck in pytest
        plt.ion()

    try:
        if verbose:
            printm('>>> _test_media_line_shift')

        setup_test_line_databases()  # add HITRAN-CO-TEST in ~/.radis if not there

        sf = SpectrumFactory(wavelength_min=4500, wavelength_max=4600,
                             wstep=0.001,
                             parallel=False, bplot=False, cutoff=1e-30,
                             path_length=0.1, mole_fraction=400e-6,
                             isotope=[1], db_use_cached=True,
                             medium='vacuum',
                             broadening_max_width=10,
                             verbose=verbose)
        sf.warnings['MissingSelfBroadeningWarning'] = 'ignore'
        sf.load_databank('HITRAN-CO-TEST')

        # Calculate in Vacuum
        sv = sf.eq_spectrum(2000)
        assert sv.conditions['medium'] == 'vacuum'

        # Calculate in Air
        # change input (not recommended to access this directly!)
        sf.input.medium = 'air'
        sa = sf.eq_spectrum(2000)
        assert sa.conditions['medium'] == 'air'

        # Compare
        if plot:
            fig = plt.figure(fig_prefix+'Propagating media line shift')
            sv.plot('radiance_noslit', nfig=fig.number, lw=2, label='Vacuum')
            plt.title('CO spectrum (2000 K)')
            sa.plot('radiance_noslit', nfig=fig.number,
                    lw=2, color='r', label='Air')

        # ... there should be about ~1.25 nm shift at 4.5 µm:
        assert np.isclose(sv.get('radiance_noslit', wunit='nm')[0][0] -
                          sa.get('radiance_noslit', wunit='nm')[0][0],
                          1.2540436086346745)

    except DatabankNotFound as err:
        assert IgnoreMissingDatabase(err, __file__, warnings)


def _run_testcases(plot=True, verbose=True, *args, **kwargs):

    # Test power density
    test_power_integral(verbose=verbose, *args, **kwargs)

#    # Show media line_shift
    test_media_line_shift(plot=plot, verbose=verbose, *args, **kwargs)

    # Test spectrum generation
    test_spec_generation(plot=plot, verbose=2*True, *args, **kwargs)

    return True


# --------------------------
if __name__ == '__main__':
    printm('Testing factory:', _run_testcases(verbose=True))



# -*- coding: utf-8 -*-
"""
Created on Mon Nov 20 10:59:49 2017

@author: erwan

Examples
--------

Run all tests::

    pytest       (in command line, in project folder)

Run only fast tests (i.e: tests that a 'fast' label)::

    pytest -m fast


-------------------------------------------------------------------------------


"""

from __future__ import unicode_literals, print_function, absolute_import, division
import os
import matplotlib.pyplot as plt
import sys
from radis.lbl import SpectrumFactory
from radis.misc.printer import printm
from radis.tools.database import load_spec
from radis.misc.utils import DatabankNotFound
from radis.test.utils import IgnoreMissingDatabase, setup_test_line_databases
from os.path import basename, exists
fig_prefix = basename(__file__)+': '

# %% Test


def test_load_spectrum(plot=False, verbose=True, warnings=True, *args, **kwargs):
    ''' Test load / save 

    Fast version: dont save lines / populations, compare spectra only 
    '''
    
    setup_test_line_databases()

    temp_file_name = '_test_database_co2_tempfile.spec'
    assert(not exists(temp_file_name))

    try:
        sf = SpectrumFactory(
            wavelength_min=4190,
            wavelength_max=4200,
            mole_fraction=400e-6,
            path_length=0.1,  # cm
            isotope=[1],
            db_use_cached=True,
            cutoff=1e-20,
            verbose=verbose)
        sf.warnings['MissingSelfBroadeningWarning'] = 'ignore'
        sf.load_databank('HITRAN-CO2-TEST')
        s1 = sf.eq_spectrum(Tgas=300)
        s1.apply_slit(2, verbose=False)
        s1.update()
        s2 = load_spec(s1.store(temp_file_name,
                                compress=True))
        s2.update()
        if plot:
            fig = plt.figure(fig_prefix+'Calculated vs stored+retrieved')
            s1.plot('absorbance', nfig=fig.number, lw=3, label='calculated')
            s2.plot('absorbance', nfig=fig.number, color='r',
                    label='stored (compressed) and retrieved')
            plt.legend()

        assert s1.compare_with(s2, spectra_only=True, plot=False)

    except DatabankNotFound as err:
        assert IgnoreMissingDatabase(err, __file__, warnings)

    finally:  # cleaning
        if exists(temp_file_name):
            os.remove(temp_file_name)

    return True


def test_load_lines_pops(plot=False, verbose=True, warnings=True, *args, **kwargs):
    ''' Test load / save 

    Full version: save lines, populations (hundreds of MB for CO2, much less
    for CO), compare everything 
    '''

    temp_file_name = '_test_database_co_tempfile.spec'
    assert(not exists(temp_file_name))

    try:
        sf = SpectrumFactory(
            wavelength_min=4500,
            wavelength_max=4800,
            mole_fraction=400e-6,
            path_length=0.1,  # cm
            isotope=[1, 2],
            db_use_cached=True,
            cutoff=1e-20,
            verbose=verbose)
        sf.warnings['MissingSelfBroadeningWarning'] = 'ignore'
        sf.load_databank('HITRAN-CO-TEST')
        s1 = sf.non_eq_spectrum(Tvib=300, Trot=300)
        s1.apply_slit(2, verbose=False)
        s1.update()
        s2 = load_spec(s1.store(temp_file_name,
                                discard=[],   # we're actually saving lines and
                                              # populations here! expect hundreds
                                              # of megabytes for big molecules
                                              # :line CO2 ... here CO is fine
                                              # (by default lines and populations
                                              # are discarded)
                                compress=True    # only removes some spectral
                                                 # quantities, cant change population
                                                 # or lines size
                                ),
                       binary=True,     # uncompress
                       )
        s2.update()

        assert s1.compare_with(s2, spectra_only=False, plot=False)

        # Test with json-tricks directly
        # Does not work yet
#        from json_tricks import dumps, loads
#        s2b = loads(dumps(s1))
#        assert s1.compare_with(s2b, spectra_only=False, plot=False)

    except DatabankNotFound as err:
        assert IgnoreMissingDatabase(err, __file__, warnings)

    finally:  # cleaning
        if exists(temp_file_name):
            os.remove(temp_file_name)

    return True


def _run_testcases(plot=False, verbose=True, warnings=True, *args, **kwargs):

    test_load_spectrum(plot=plot, verbose=verbose, warnings=warnings,
                       *args, **kwargs)

    test_load_lines_pops(plot=plot, verbose=verbose, warnings=warnings,
                         *args, **kwargs)

    return True


if __name__ == '__main__':

    printm('Testing database.py: ', _run_testcases(plot=True))



# -*- coding: utf-8 -*-
"""

-------------------------------------------------------------------------------

"""

from __future__ import absolute_import, division, print_function, unicode_literals

from .hitran import hit2df, get_molecule, get_molecule_identifier
from .cdsd import cdsd2df


# %% Get list of supported molecules
def _get_supported_molecules_equilibrium():
    ''' Molecules supported in RADIS equilibrium calculations
    Basically, all [HITRAN-2016]_ species
    '''
    # Add all HITRAN species
    from radis.io.hitran import (HITRAN_CLASS1, HITRAN_CLASS2, HITRAN_CLASS3,
                                 HITRAN_CLASS4, HITRAN_CLASS5, HITRAN_CLASS6,
                                 HITRAN_CLASS7, HITRAN_CLASS8, HITRAN_CLASS9,
                                 HITRAN_CLASS10)

    return (HITRAN_CLASS1+HITRAN_CLASS2+HITRAN_CLASS3+HITRAN_CLASS4+HITRAN_CLASS5 +
            HITRAN_CLASS6+HITRAN_CLASS7+HITRAN_CLASS8+HITRAN_CLASS9+HITRAN_CLASS10)


def _get_supported_molecules_nonequilibrium():
    ''' Molecules supported in RADIS non equilibrium calculations without need
    for extra databases. 
    Basically, molecules whose spectroscopic constants are built-in RADIS database
    (see radis.db)
    '''

    from radis.db.molecules import Molecules

    return list(Molecules.keys())


MOLECULES_LIST_EQUILIBRIUM = __supported_molecules_equilibrium__ = _get_supported_molecules_equilibrium()
''' list: molecules that can be calculated in RADIS at equilibrium. 
Basically, all [HITRAN-2016]_ species are available. '''
MOLECULES_LIST_NONEQUILIBRIUM = __supported_molecules_nonequilibrium__ = _get_supported_molecules_nonequilibrium()
''' list: molecules that can be calculated in RADIS at nonequilibrium. Built-in
spectroscopic constants to calculate energy levels are needed 

See Also
--------

:py:data:`~radis.db.molecules.Molecules`,
:py:func:`~radis.db.molecules.getMolecule`
'''



# -*- coding: utf-8 -*-
"""
Created on Thu Sep 14 13:44:35 2017

@author: erwan

Summary
-------

An experimental module to merge precalculated vibrational bands in post-processing,
and recompute a new spectrum at a new temperature or with overpopulations, without
having to recalculate the broadening of each line

.. warning::

    Order of magnitude faster, but only valid under optically thin conditions as the
    rescaling of absorption doesnt scale induced emission properly

----------

"""

from __future__ import absolute_import
from __future__ import print_function
from warnings import warn
from radis.spectrum.spectrum import is_spectrum
from radis.spectrum.rescale import (rescale_abscoeff, rescale_absorbance, rescale_emisscoeff,
                                    rescale_radiance_noslit, rescale_transmittance_noslit)
from radis.spectrum.utils import NON_CONVOLUTED_QUANTITIES
from radis.misc.basics import is_float
from radis.phys.constants import hc_k
from radis.lbl.labels import (vib_lvl_name_cdsd_pc, vib_lvl_name_cdsd_pcN)
import numpy as np
from numpy import exp
from time import time
import pandas as pd

# keys that may not be equals for different bands
_IGNORE_KEYS = ['band', 'band_htrn', 'viblvl_u', 'viblvl_l']


class LevelsList(object):
    ''' A class to generate a Spectrum from a list of precalculated bands at 
    a given reference temperature 
    
    .. warning::
        
        only valid under optically thin conditions!!
    
    '''

    # hardcode attribute names, but can save a lot of memory
    __slots__ = ['Tvib_ref', 'Trot_ref', 'vib_distribution_ref',
                 'mole_fraction_ref', 'path_length_ref',
                 'sorted_bands', 'cum_weight', 'E_bands',
                 'bands_ref', 'verbose', 'bands', 'parfunc', 'levelsfmt',
                 'lvl_index', 'vib_levels', 'copy_lines', 'Qref', 'Qvib_ref']

    def __init__(self, parfunc, bands, levelsfmt, sortby='Ei', copy_lines=False, verbose=True):
        ''' 
        Parameters
        ----------

        bands: dict of bands
            bands are Spectrum objects calculated at equilibrium or non-equilibrim. 
            Tgas, or (Tvib, Trot) must be given and the same in all bands conditions. 

        '''

        # Check inputs
        for br, s in bands.items():
            if not is_spectrum(s):
                raise ValueError('`bands` must be a list of Spectrum objects. ' +
                                 'Got {0}'.format(type(s)))

        # Assert all conditions are the same
        bands_list = list(bands.values())
        cond_ref = bands_list[0].conditions
        for s in bands_list[1:]:   # type(s): Spectrum
            if list(s.conditions.keys()) != list(cond_ref.keys()):
                raise ValueError(
                    'Input conditions must be the same for all bands')
            for k in cond_ref.keys():
                if k in _IGNORE_KEYS:   # don't compare these entries
                    continue
                if s.conditions[k] != cond_ref[k]:
                    raise ValueError('Input conditions must be the same for all bands' +
                                     '. Got a difference for entry {0}: {1} vs {2}'.format(
                                         k, s.conditions[k], cond_ref[k]))

        if 'Tvib' in cond_ref:
            Tvib_ref = cond_ref['Tvib']
        else:
            Tvib_ref = cond_ref['Tgas']
        if 'Trot' in cond_ref:
            Trot_ref = cond_ref['Trot']
        else:
            Trot_ref = cond_ref['Tgas']
        vib_distribution_ref = cond_ref['vib_distribution']
        if 'overpopulation' in cond_ref:
            raise NotImplementedError

        # stores input conditions
        self.Tvib_ref = Tvib_ref
        self.Trot_ref = Trot_ref
        self.mole_fraction_ref = cond_ref['mole_fraction']
        self.path_length_ref = cond_ref['path_length']
        self.vib_distribution_ref = vib_distribution_ref

        # stores computation parameters
        # TODO: store it in parfunc instead (and make parfunc an EnergyDatabase)
        self.levelsfmt = levelsfmt

        # %% Misc
        self.verbose = verbose
        self.copy_lines = copy_lines

        # %% Set up energies  ( use a partition function object for the moment)

        self.parfunc = parfunc
        self._init_levels(sortby)
        self._connect_bands_to_levels(bands)

        Qvib = self._calc_vib_populations(Tvib=self.Tvib_ref,
                                          vib_distribution=self.vib_distribution_ref)

        self.Qref = parfunc.at_noneq(Tvib_ref, Trot_ref,
                                     vib_distribution=vib_distribution_ref)  # TODO: add overpopulations in partition function
        self.Qvib_ref = Qvib

        self.vib_levels['nvib_ref'] = self.vib_levels['nvib']
        del self.vib_levels['nvib']

        # %% Set up bands
        self._init_bands_ref(bands, sortby)

    def _init_bands_ref(self, bands, sortby):

        # %% Parse bands

        #        # Get band names   (if bands was a list)
        #        band_dict = {}
        #        for bd in bands:
        #            try:
        #                band_name = set(bd.lines['band'])
        #            except:
        #                raise ValueError('Bands must have been created with a `band` attribute'+\
        #                                 ' in lines')
        #            assert len(band_name) == 0
        #            band_dict[band_name[0]] = bd
        #        self.bands = band_dict
        #        self.bands0 = {br: bands[br].copy(copy_lines=False) for br in bands}  # make a copy

        # Get energy of upper level of band
        E_bands = {}
        for br, band in bands.items():               # type(band): Spectrum
            if br == 'others':
                continue
            E_bands[br] = np.mean(list(set(band.lines['Evibu'])))
            # Todo: should be all equals ?
            assert len(set(band.lines['Evibu'])) == 1

        # Calculate band weight
        weight = {}
        for br, band in bands.items():       # type(band): Spectrum
            weight[br] = band.lines[sortby].sum()
        tot = np.sum([np.array(w) for w in weight.values()])
        for br, band in bands.items():
            weight[br] /= tot
        sorted_bands = sorted(bands, key=lambda br: weight[br])[::-1]
        cum_weight = np.array([weight[br] for br in sorted_bands])
        cum_weight = np.cumsum(cum_weight)

        # Store
        self.sorted_bands = sorted_bands
        self.cum_weight = cum_weight
        self.E_bands = E_bands

#        # Delete lines (makes copies faster)
#        for br in self.bands0:
#            del self.bands0[br].lines

        # Initialize bands
        self.bands_ref = {br: band.copy(
            copy_lines=self.copy_lines) for br, band in bands.items()}

    def _init_levels(self, sortby):
        ''' 
        Notes
        -----

        vib level in (p,c,j,N) notation is ambiguous. Here we use (p,c) based
        on energy difference and perturbation rules
        '''

        df = self.parfunc.df
        levelsfmt = self.levelsfmt

        if levelsfmt == 'cdsd-pc':
            vib_index = ['p', 'c']
            def vib_lvl_name_cdsd(p, c, N): 
                return vib_lvl_name_cdsd_pc(p, c)
            
        elif levelsfmt == 'cdsd-pcN':
            vib_index = ['p', 'c', 'N']
            vib_lvl_name_cdsd = vib_lvl_name_cdsd_pcN

        else:
            raise NotImplementedError(levelsfmt)

        vib_levels = df.drop_duplicates(subset=vib_index)  # 192 ms ± 1.13 ms

        # Add levels
        def add_viblvl(row):
            '''  '''
        #    row['vib_lvl'] = '({p},{c},{N})'.format(**dict([(k,int(row[k])) for k in ['p', 'c', 'N']]))
#            row['viblvl'] = '({p},{c})'.format(**dict([(k,int(row[k])) for k in ['p', 'c']]))
            row['viblvl'] = vib_lvl_name_cdsd(row.p, row.c, row.N)
            return row

        t0 = time()
        vib_levels = vib_levels.apply(add_viblvl, axis=1)
        print('Added viblvl in {0:.1f}s'.format(time()-t0))

        #vib_levels = vib_levels.set_index(['p', 'c', 'N'])
        vib_levels = vib_levels.set_index(['viblvl'])

        # %% Generate index of  levels

        lvl_index = dict().fromkeys(list(vib_levels.index))
        for k in lvl_index:  # Note: do not fill directly in frmokeys or you'll get the same shared list
            lvl_index[k] = {'bands_where_low': [],
                            'bands_where_up': []}

        self.lvl_index = lvl_index
        self.vib_levels = vib_levels

    def _connect_bands_to_levels(self, s_bands):

        index = self.lvl_index

        for s in s_bands.values():
            if s.name == 'others':
                continue

            viblvl_u = s.lines['viblvl_u'].iloc[0]
            viblvl_l = s.lines['viblvl_l'].iloc[0]

            index[viblvl_u]['bands_where_up'].append(s)
            index[viblvl_l]['bands_where_low'].append(s)

    def plot_vib_populations(self, nfig=None, **kwargs):
        ''' Plot current distribution of vibrational levels

        By constructions populations are shown as divided by the state degeneracy,
        i.e,            g =   gv   *   (2J+1)   *    gi    *     gs

        Parameters
        ----------

        nfig: str, int
            name of Figure to plot on 

        kwargs: **dict
            arguments are forwarded to plot()

        '''

        import matplotlib.pyplot as plt

        vib_levels = self.vib_levels

        plt.figure(num=nfig)
        plt.plot(vib_levels.E, vib_levels.nvib, 'ok', **kwargs)
        plt.xlabel('Energy (cm-1)')
        plt.ylabel('Population')
        plt.yscale('log')
        plt.xlim(xmin=0)

        return

    def _calc_vib_populations(self, Tvib, vib_distribution='boltzmann',
                              overpopulation=None):
        ''' Calculate vibrational populations from Tref to new Tvib and store
        results in vib_levels dataframe
        This does not modify the spectra yet! 

        By constructions populations are calculated divided by the state degeneracy,
        i.e,            g =   gv   *   (2J+1)   *    gi    *     gs
        This means we should only use ratios for rescaling 

        The information on state degeneracy and isotopic abundance is already
        included in the linestrength / emission integral, hence in the pre-calculated 
        emisscoeff / abscoeff

        '''

        vib_levels = self.vib_levels
        if overpopulation is None:
            overpopulation = {}

        if is_float(Tvib):

            # Get new population
            E_vib = vib_levels['Evib']
            g = 1   # explicitely calculate populations divided by degeneracy
            # this means we should only use ratios for rescaling
            if vib_distribution == 'boltzmann':
                nvibQvib = g * exp(-hc_k*E_vib/Tvib)
            else:
                raise NotImplementedError(
                    'vib_distribution: {0}'.format(vib_distribution))

        else:
            Tvib1, Tvib2, Tvib3 = Tvib

            # Get new population
            E_vib1 = vib_levels['Evib1']
            E_vib2 = vib_levels['Evib2']
            E_vib3 = vib_levels['Evib3']
            g = 1   # explicitely calculate populations divided by degeneracy
            # this means we should only use ratios for rescaling
            if vib_distribution == 'boltzmann':
                nvibQvib = g * exp(-hc_k*E_vib1/Tvib1) * \
                    exp(-hc_k*E_vib2/Tvib2) * exp(-hc_k*E_vib3/Tvib3)
            else:
                raise NotImplementedError(
                    'vib_distribution: {0}'.format(vib_distribution))

        # Add overpopulation
        if overpopulation != {}:
            for k, ov in overpopulation.items():
                nvibQvib.loc[k] *= ov
                #TODO: test
            warn('NotImplemented: partition function overpopulation correction not tested')

        # Normalize with partition function
        Qvib = nvibQvib.sum()
        nvib = nvibQvib/Qvib

        # update dataframe
        vib_levels['nvib'] = nvib
        vib_levels['Qvib'] = Qvib

        return Qvib

#    def set_cutoff(self, cutoff=1e-3):
#        ''' Discard most of the bands and keep error below cutoff
#
#        Warning: this cannot be undone!'''
#
#        cum_weight = self.cum_weight
#        sorted_bands = self.sorted_bands
#
#        cutoff_i = np.argmax((1-cum_weight)<cutoff)
#        first_bands = sorted_bands[:cutoff_i+1]
#
#        # Update bands
#        new_bands = {br: self.bands_ref[br] for br in first_bands}
#
#        if self.verbose:
#            print(('Set cutoff. Discarded {0}/{1} bands. Estimated error: {2:.2f}%'.format(
#                len(self.bands_ref)-len(new_bands), len(self.bands_ref),
#                (1-cum_weight[cutoff_i])*100)))
#
#        self.bands_ref = new_bands

    def eq_spectrum(self, Tgas, overpopulation=None, mole_fraction=None, path_length=None,
                    save_rescaled_bands=False):
        ''' See :py:meth:`~radis.lbl.factory.SpectrumFactory.eq_spectrum`
        
        .. warning::
            
            only valid under optically thin conditions!!
        
        Parameters
        ----------

        ... same as usually. If None, then the reference value (used to 
        calculate bands) is used 

        Other Parameters
        ----------------

        save_rescaled_bands: boolean
            save updated bands. Take some time as it requires rescaling all 
            bands individually (which is only done on the MergedSlabs usually)
            Default ``False`` 

        '''
        s = self.non_eq_spectrum(Tgas, Tgas, overpopulation=overpopulation,
                                 mole_fraction=mole_fraction, path_length=path_length,
                                 save_rescaled_bands=save_rescaled_bands)
        del s.conditions['Tvib']
        del s.conditions['Trot']
        s.conditions['Tgas'] = Tgas

        return s

    def non_eq_spectrum(self, Tvib=None, Trot=None, Ttrans=None,
                        vib_distribution='boltzmann', overpopulation=None,
                        mole_fraction=None, path_length=None,
                        save_rescaled_bands=False):
        ''' See :py:meth:`~radis.lbl.factory.SpectrumFactory.non_eq_spectrum`
        
        .. warning::
            
            only valid under optically thin conditions!!
        
        Parameters
        ----------

        ... same as usually. If None, then the reference value (used to 
        calculate bands) is used 

        Other Parameters
        ----------------

        save_rescaled_bands: boolean
            save updated bands. Take some time as it requires rescaling all 
            bands individually (which is only done on the MergedSlabs usually)
            Default ``False`` 

        Notes
        -----

        Implementation:

        Generation of a new spectrum is done by recombination of the precalculated
        bands with 

        '''

        from radis.los import MergeSlabs

        # Restart from a copy each time  (else reference bands are modified by rescaling
        # and we may loose information if rescaling to 0 for instance)
        bands_ref = self.bands_ref
        bands = {br: bands_ref[br].copy(
            copy_lines=self.copy_lines) for br in bands_ref}

        # Initialize inputs
        if Trot is None:
            Trot = self.Trot_ref
        if Tvib is None:
            Tvib = self.Tvib_ref
        if Trot != self.Trot_ref:
            raise ValueError('Trot {0} doesnt match the reference Trot: {1}'.format(Trot,
                                                                                    self.Trot_ref))
        if mole_fraction is None:
            mole_fraction = self.mole_fraction_ref
        if path_length is None:
            path_length = self.path_length_ref
        if overpopulation is None:
            overpopulation = {}

#        # Fill missing overpopulation with 1
#        if overpopulation is None:
#            overpopulation = dict.fromkeys(list(bands.keys()), 1)
#        else:
#            for br in bands.keys():
#                if not br in overpopulation:
#                    overpopulation[br] = 1

#        Tvib_ref = self.Tvib_ref
#        pop_correction = dict.fromkeys(list(bands.keys()), 1)
#        if Tvib != Tvib_ref:
#            raise NotImplementedError
#            E_bands = self.E_bands
#            # Correct for vibrational temperature
#            for br in bands:
#                pop_correction[br] = exp(-E_bands[br]*hc_k/Tvib)/exp(-E_bands[br]*hc_k/Tvib_ref)

        # Recalculate populations from reference everytime
        vib_levels = self.vib_levels

        # Recalculate partition function
        Qvib = self._calc_vib_populations(Tvib=Tvib, vib_distribution=vib_distribution,
                                          overpopulation=overpopulation)
        # TODO: note: Qvib doesnt take gvib into account. Wrong Qvib but correct populations
        # if rescaled with ratio???
        # ... warning: update_populations set to False here not updated (because we dont
        # ... care much about all levels here. only the one from vib_levels matter)
        if overpopulation is not None:
            Q, _, _ = self.parfunc.at_noneq(Tvib, Trot, overpopulation=overpopulation,
                                            returnQvibQrot=True, update_populations=False)
        else:
            Q = self.parfunc.at_noneq(Tvib, Trot, update_populations=False)

        Qref = self.Qref
        Qvib_ref = self.Qvib_ref

        for br, band in bands.items():         # type(band): Spectrum
            if br == 'others':
                continue
            viblvl_u = band.conditions['viblvl_u']
            viblvl_l = band.conditions['viblvl_l']
            nu_vib = vib_levels.loc[viblvl_u, 'nvib']
            nu_vib_old = vib_levels.loc[viblvl_u, 'nvib_ref']
            nl_vib = vib_levels.loc[viblvl_l, 'nvib']
            nl_vib_old = vib_levels.loc[viblvl_l, 'nvib_ref']

            # with new, old = (2, 1):
            # n2/n1 = exp(-Evib2/kT2)/exp(-Evib1/kT1) * Qvib1/Qvib2
            #       = n2vib / n1vib * Q1/Q2 * Q2vib / Q1vib
            if vib_distribution == 'boltzmann':
                corfactor_u = nu_vib/nu_vib_old * Qref/Q * Qvib/Qvib_ref
                corfactor_l = nl_vib/nl_vib_old * Qref/Q * Qvib/Qvib_ref
            else:
                raise NotImplementedError(
                    'vib_distribution: {0}'.format(vib_distribution))

            rescale_updown_levels(band, corfactor_u, 1, corfactor_l, 1)

            # Update lines
            if self.copy_lines:
                band.lines['nu'] *= corfactor_u
                band.lines['nl'] *= corfactor_l
                band.lines['Qvib'] = Qvib
                band.lines['nu_vib'] *= corfactor_u
                band.lines['nl_vib'] *= corfactor_l
                band.lines['Ei'] *= np.nan
                band.lines['S'] = np.nan

        # Get total spectrum
        s = MergeSlabs(*list(bands.values()))

        # populations
        s.populations = pd.DataFrame(vib_levels[['nvib', 'Evib']])

        # rebuild lines
        if self.copy_lines:
            s.lines = pd.concat([band.lines for band in bands.values()])

        # Update total mole fraction if it changed
        mole_fraction_ref = self.mole_fraction_ref
        if mole_fraction != mole_fraction_ref:
            s.rescale_mole_fraction(mole_fraction, mole_fraction_ref,
                                    ignore_warnings=True,  # mole_fraction is 'N/A'
                                    force=True)

        path_length_ref = self.path_length_ref
        if path_length != path_length_ref:
            s.rescale_path_length(path_length, path_length_ref, force=True)

        # Add parameters in conditions:
        s.conditions['overpopulation'] = overpopulation
        s.conditions['mole_fraction'] = mole_fraction  # was 'N/A' after Merge
        # because it's different for all bands
        s.conditions['path_length'] = path_length
        s.conditions['Tvib'] = Tvib
        s.conditions['Trot'] = Trot

        if save_rescaled_bands:
            for br, band in bands.items():          # type(band): Spectrum
                if mole_fraction != mole_fraction_ref:
                    band.rescale_mole_fraction(mole_fraction, mole_fraction_ref,
                                               ignore_warnings=True,  # mole_fraction is 'N/A'
                                               force=True)
                if path_length != path_length_ref:
                    band.rescale_path_length(
                        path_length, path_length_ref, force=True)
            self.bands = bands

        return s


# %% Overpopulation functions

def rescale_updown_levels(spec, new_nu, old_nu, new_nl, old_nl,
                          ignore_warnings=False, force=False):
    ''' Update spectrum with new molar fraction for upper and lower levels

    Convoluted values (with slit) are dropped in the process.

    Rescales with a ratio new_nu/old_nu. 

    This is only valid for emission quantities, under optically thin conditions,
    as rescaling doesnt correct for induced emission. 

    Parameters
    ----------

    new_nu: float
        new upper state mole fraction

    old_nu: float
        current upper state mole fraction

    new_nl: float
        new lower state mole fraction

    old_nl: float
        current lower state mole fraction

    Other Parameters
    ----------------

    force: boolean
        if False, won't allow rescaling to 0 (not to loose information).
        Default ``False``

    Notes
    -----

    Implementation:

    similar to rescale_path_length() but we have to scale abscoeff & emisscoeff
    Note that this is valid only for small changes in mole fractions. Then,
    the change in line broadening becomes significant and the whole band
    should be recomputed

    IMPORTANT: if editing make sure you use the proper nu and nl. In particular
    when infering emission quantities from absorption quantities this may
    ends up in error in overpopulation rescaling.

    # TODO: Add warning when too large rescaling
    '''

    # Check inputs
    # ---------
#        if old_mole_fraction is not None:
#            try:
#                if spec.conditions['mole_fraction'] != old_mole_fraction and not ignore_warnings:
#                    warn('mole_fraction ({0}) doesnt match value given in conditions ({1})'.format(
#                            old_mole_fraction, spec.conditions['mole_fraction']))
#            except KeyError: # mole fraction not defined
#                pass

#        else:
#            try:
#                old_mole_fraction = spec.conditions['mole_fraction']
#            except KeyError:
#                raise KeyError('mole_fraction has to be defined in conditions (or use'+\
#                                ' `from_mole_fraction`)'),

    if (new_nu < 0 or new_nl < 0) and not force:
        raise ValueError('mole_fraction cannot be negative')
    if (new_nu == 0 or new_nl == 0) and not force:
        raise ValueError('Rescaling to 0 will loose information. Choose force '
                         '= True')

    for q in ['transmittance', 'radiance']:
        qns = q+'_noslit'
        qties = spec.get_vars()
        if q in qties and qns not in qties and not force:
            raise KeyError('Cant rescale {0} if {1} not stored'.format(q, qns) +
                           ' Use force=True to rescale anyway. {0}'.format(q) +
                           ' will be deleted')

    # Get path length
    if 'path_length' in list(spec.conditions.keys()):
        path_length = spec.conditions['path_length']
        true_path_length = True
    else:
        path_length = 1
        true_path_length = False

    # %%  Rescale

    optically_thin = spec.is_optically_thin()

    # Choose which values to recompute
    # ----------
    initial = spec.get_vars()               # quantities initialy in spectrum
    recompute = list(initial)               # quantities to recompute
    rescaled = {}                           # quantities rescaled

    if ('radiance_noslit' in initial and not optically_thin):
        recompute.append('emisscoeff')
    if ('absorbance' in initial or 'transmittance_noslit' in initial
            or 'radiance_noslit' in initial and not optically_thin):
        recompute.append('abscoeff')
    extra = []
    recompute = set(recompute)  # remove duplicates

    # Get units
    units = spec.units.copy()

    # Recompute!
    # ----------
    waveunit = spec.get_waveunit()  # keep all quantities in same waveunit

    if 'abscoeff' in recompute:
        rescaled, units = rescale_abscoeff(spec, rescaled, initial,     # TODO: remove rescaled = ... Dict is mutable
                                           old_nl, new_nl, path_length,
                                           waveunit, units, extra, true_path_length)

    if 'emisscoeff' in recompute:
        rescaled, units = rescale_emisscoeff(spec, rescaled, initial,
                                             old_nu, new_nu, path_length,
                                             optically_thin, waveunit, units, extra, true_path_length)

    if 'absorbance' in recompute:
        rescaled, units = rescale_absorbance(spec, rescaled, initial,
                                             old_nl, new_nl, path_length,
                                             path_length, waveunit, units, extra, true_path_length)

    if 'transmittance_noslit' in recompute:
        rescaled, units = rescale_transmittance_noslit(spec, rescaled, initial,
                                                       old_nl, new_nl, path_length,
                                                       path_length, waveunit, units, extra,
                                                       true_path_length)

    if 'radiance_noslit' in recompute:
        rescaled, units = rescale_radiance_noslit(spec, rescaled, initial,
                                                  old_nu, new_nu, path_length,
                                                  path_length, optically_thin, waveunit, units,
                                                  extra, true_path_length)

    # Save (only) the ones that were in the spectrum initially
    for q in rescaled:
        if q in initial:
            if q in NON_CONVOLUTED_QUANTITIES:
                spec._q[q] = rescaled[q]
            else:
                spec._q_conv[q] = rescaled[q]

    # Update units
    for k, u in units.items():
        spec.units[k] = u

    # Drop convoluted values
    for q in ['transmittance', 'radiance']:
        if q in list(spec._q_conv.keys()):
            del spec._q_conv[q]

    # Reapply slit if possible
    if ('slit_function' in spec.conditions and 'slit_unit' in spec.conditions
            and 'norm_by' in spec.conditions):
        slit_function = spec.conditions['slit_function']
        slit_unit = spec.conditions['slit_unit']
        norm_by = spec.conditions['norm_by']
        try:
            shape = spec.conditions['shape']
        except KeyError:
            shape = 'triangular'
        spec.apply_slit(slit_function=slit_function, unit=slit_unit, shape=shape,
                        norm_by=norm_by)

    # Update conditions
    spec.conditions['mole_fraction'] = new_nl if new_nl == new_nu else 'N/A'


def _test(verbose=True, plot=False, *args, **kwargs):
    ''' Generate scalable LevelList at 1500K. Rescale at 2000K and compare with
    spectrum directly calculated at 2000K 
    '''

    from radis.lbl import SpectrumFactory

    iso = 1
    sf = SpectrumFactory(
        wavelength_min=4170,
        wavelength_max=4200,
        mole_fraction=1,
        path_length=0.05,
        cutoff=1e-25,
        #                     isotope=[1,2],
        isotope=iso,
        db_use_cached=True,
        wstep=0.01,
        broadening_max_width=10,
        medium='air',
        verbose=verbose)

    sf.load_databank('CDSD')

    parfunc = sf.parsum_calc['CO2'][iso]['X']

    # %% Fill levels for all bands

    Tref = 1500
    s_bands = sf.non_eq_bands(Tvib=Tref, Trot=Tref)
    lvlist = LevelsList(parfunc, s_bands, sf.params.levelsfmt)

    # %% Test
    Tvib = 2000
    s_resc = lvlist.non_eq_spectrum(Tvib=Tvib, Trot=Tref)
    s0 = sf.non_eq_spectrum(Tvib, Tref)

    return s0.compare_with(s_resc, spectra_only=True, plot=plot)


if __name__ == '__main__':

    from radis.test.lbl.test_overp import run_testcases
    print('test_overp.py:', run_testcases(plot=True))



# -*- coding: utf-8 -*-
"""

Predefined molecules using RADIS built-in
:ref:`spectroscopic constants <label_db_spectroscopic_constants>`

See :py:data:`~radis.db.molecules.Molecules`

-------------------------------------------------------------------------------


"""

from __future__ import division, absolute_import, print_function, unicode_literals
from radis.db.classes import ElectronicState
from radis.db.utils import get_dunham_coefficients, get_herzberg_coefficients


# %% Define some commonly used molecules

from radis.phys.convert import eV2cm

# CO
# ----------

# Define with default diatomic constants
CO_X_iso1 = ElectronicState('CO', isotope=1, state='X', term_symbol='1Σ+',
                            spectroscopic_constants=get_dunham_coefficients('CO', 1, 'X1SIG+'),
                            vmax=17,        # max level for Dunham's expansion
                            vmax_morse=48,
                            Ediss=eV2cm(11.16),
                            )
CO_X_iso2 = ElectronicState('CO', isotope=2, state='X', term_symbol='1Σ+',
                            spectroscopic_constants=get_dunham_coefficients('CO', 2, 'X1SIG+'),
                            vmax=17,        # max level for Dunham's expansion
                            vmax_morse=48,
                            Ediss=eV2cm(11.16),
                            )
CO_X_iso3 = ElectronicState('CO', isotope=3, state='X', term_symbol='1Σ+',
                            spectroscopic_constants=get_dunham_coefficients('CO', 3, 'X1SIG+'),
                            vmax=17,        # max level for Dunham's expansion
                            vmax_morse=48,
                            Ediss=eV2cm(11.16),
                            )

# CO2
# ----------

# Generate the energy

#from radis.db.CO2.energies import E_CO2_626_X1SIGg, Ehaj123_CO2_626_X1SIGg
from radis.levels.vibrating_rotor import EvJ_uncoupled_vibrating_rotor, EvJah_uncoupled_vibrating_rotor

#def build_CO2_X(isotope):
#    coeff_dict =get_herzberg_coefficients('CO2', isotope, 'X1SIGu+')
#    # Update standard vibrating rotor formula with CO2 specific generacies and 
#    # Herzberb coeffs
#    def Erovib_CO2(v1, v2, l2, v3, J, remove_ZPE=True):
#        ''' Rovibrational energies.
#        See :py:func:`~radis.levels.vibrating_rotor.EvJ_uncoupled_vibrating_rotor` 
#        '''
#        return EvJ_uncoupled_vibrating_rotor(v1, v2, l2, v3, J, coeff_dict=coeff_dict,
#                                             gv1=1, gv2=2, gv3=1, remove_ZPE=remove_ZPE)
#    def Ehaj_CO2(v1, v2, l2, v3, J, remove_ZPE=True):
#        ''' Harmonic and anharmonic parts of rovibrational energies.
#        See :py:func:`~radis.levels.vibrating_rotor.EvJah_uncoupled_vibrating_rotor` 
#        '''
#        return EvJah_uncoupled_vibrating_rotor(v1, v2, l2, v3, J, coeff_dict=coeff_dict,
#                                             remove_ZPE=remove_ZPE)
#    return ElectronicState('CO2', isotope=1, state='X', term_symbol='1Σu+',
##                            spectroscopic_constants=get_herzberg_coefficients('CO2', 1, 'X1SIGu+'),
#                            Erovib=Erovib_CO2,
#                            Ehaj=Ehaj_CO2,  # for Treanor
#                            Ediss=44600,
#                            )


CO2_X_626 = ElectronicState('CO2', isotope=1, state='X', term_symbol='1Σu+',
                            spectroscopic_constants=get_herzberg_coefficients('CO2', 1, 'X1SIGu+'),
#                            Erovib=EvJah_uncoupled_vibrating_rotor,
#                            Ehaj=Ehaj_CO2,  # for Treanor
                            Ediss=44600,
                            )
CO2_X_636 = ElectronicState('CO2', isotope=2, state='X', term_symbol='1Σu+',
                            spectroscopic_constants=get_herzberg_coefficients('CO2', 2, 'X1SIGu+'),
#                            Erovib=EvJah_uncoupled_vibrating_rotor,
#                            Ehaj=Ehaj_CO2,  # for Treanor
                            Ediss=44600,
                            )
CO2_X_628 = ElectronicState('CO2', isotope=3, state='X', term_symbol='1Σu+',
                            spectroscopic_constants=get_herzberg_coefficients('CO2', 3, 'X1SIGu+'),
#                            Erovib=EvJah_uncoupled_vibrating_rotor,
#                            Ehaj=Ehaj_CO2,  # for Treanor
                            Ediss=44600,
                            )
CO2_X_627 = ElectronicState('CO2', isotope=4, state='X', term_symbol='1Σu+',
                            spectroscopic_constants=get_herzberg_coefficients('CO2', 4, 'X1SIGu+'),
#                            Erovib=EvJah_uncoupled_vibrating_rotor,
#                            Ehaj=Ehaj_CO2,  # for Treanor
                            Ediss=44600,
                            )
#CO2_X_626 = build_CO2_X(1)
#CO2_X_636 = build_CO2_X(2)
#CO2_X_628 = build_CO2_X(3)
#CO2_X_627 = build_CO2_X(4)

#CO2_X_626 = ElectronicState('CO2', isotope=1, state='X', term_symbol='1Σu+',
#                            spectroscopic_constants=get_herzberg_coefficients('CO2', 1, 'X1SIGu+'),
#                            Erovib=EvJ_uncoupled_vibrating_rotor,
#                            Ehaj=EvJah_uncoupled_vibrating_rotor,  # for Treanor
#                            Ediss=44600,
#                            )
#CO2_X_636 = ElectronicState('CO2', isotope=2, state='X', term_symbol='1Σu+',
#                            spectroscopic_constants=get_herzberg_coefficients('CO2', 2, 'X1SIGu+'),
#                            Erovib=EvJ_uncoupled_vibrating_rotor,
#                            Ehaj=EvJah_uncoupled_vibrating_rotor,  # for Treanor
#                            Ediss=44600,
#                            )
#CO2_X_628 = ElectronicState('CO2', isotope=3, state='X', term_symbol='1Σu+',
#                            spectroscopic_constants=get_herzberg_coefficients('CO2', 3, 'X1SIGu+'),
#                            Erovib=EvJ_uncoupled_vibrating_rotor,
#                            Ehaj=EvJah_uncoupled_vibrating_rotor,  # for Treanor
#                            Ediss=44600,
#                            )
#CO2_X_627 = ElectronicState('CO2', isotope=4, state='X', term_symbol='1Σu+',
#                            spectroscopic_constants=get_herzberg_coefficients('CO2', 4, 'X1SIGu+'),
#                            Erovib=EvJ_uncoupled_vibrating_rotor,
#                            Ehaj=EvJah_uncoupled_vibrating_rotor,  # for Treanor
#                            Ediss=44600,
#                            )


# %% Dictionary of predefined molecules
# -----------

#  Molecule  Isotope  ElecState
#         :      :       :
Molecules = {
    'CO':   {
        1:      {
            'X': CO_X_iso1},
        2:      {
            'X': CO_X_iso2},
        3:      {
            'X': CO_X_iso3},
    },
    'CO2':  {
        1:      {
            'X': CO2_X_626},
        2:      {
            'X': CO2_X_636},
        3:      {
            'X': CO2_X_628},
        4:      {
            'X': CO2_X_627},
    },
}
'''dict: list of Electronic states whose energy levels can be calculated with RADIS
built-in :ref:`spectroscopic constants <label_db_spectroscopic_constants>`. 
For references refer to the definition of each molecule:

CO:

- :py:data:`~radis.db.molecules.CO_X_iso1`
- :py:data:`~radis.db.molecules.CO_X_iso2`
- :py:data:`~radis.db.molecules.CO_X_iso3`

CO2:
    
- :py:data:`~radis.db.molecules.CO2_X_626`
- :py:data:`~radis.db.molecules.CO2_X_636`
- :py:data:`~radis.db.molecules.CO2_X_628`
- :py:data:`~radis.db.molecules.CO2_X_627`

See Also
--------

:py:func:`~radis.db.molecules.getMolecule`

'''





def getMolecule(molecule, isotope=None, electronic_state=None, verbose=True):
    ''' Get an :py:class:`~radis.db.classes.ElectronicState` object in the 
    RADIS :py:data:`~radis.db.molecules.Molecules` list, which use the defaults
    :ref:`spectroscopic constants <label_db_spectroscopic_constants>`.

    Parameters
    ----------

    molecule: str
        molecule name

    isotope: int, or ``None``
        isotope number. if None, only one isotope must exist in database. Else, 
        an error is raised

    electronic_state: str
        if None, only one electronic state must exist in database. Else, an error 
        is raised

    verbose: boolean
        if ``True``, print which electronic state we got
        
    Returns
    -------
    
    state: ElectronicState
        an :py:class:`~radis.db.classes.ElectronicState` object. 

    See Also
    --------
    
    :py:data:`~radis.db.molecules.Molecules`

    '''

    # molecule
    try:
        mol = Molecules[molecule]
    except KeyError:
        raise KeyError('{0} is not defined in molecules with built-in '.format(molecule) +
                       'spectroscopic constants. Choose one of: {0}'.format(list(Molecules.keys())))

    # isotope
    if isotope is None:
        if len(list(mol.keys())) != 1:
            raise ValueError('Please precise which isotope among: {0}'.format(
                list(mol.keys())))
        isotope = list(mol.keys())[0]
    try:
        iso = mol[isotope]
    except KeyError:
        raise KeyError('Isotope {0} is not defined for molecule {1}. Choose one of: {2}'.format(
            isotope, molecule, list(mol.keys())))

    # electronic state
    if electronic_state is None:
        if len(list(iso.keys())) != 1:
            raise ValueError('Please choose which electronic state among: {0}'.format(
                list(iso.keys())))
        electronic_state = list(iso.keys())[0]
    try:
        state = iso[electronic_state]
    except KeyError:
        raise KeyError('{0} is not defined for molecule {1}(iso={2}). Choose one of: {3}'.format(
            electronic_state, molecule, isotope, list(mol.keys())))

    # print name
    if verbose>=2:
        print(('Found {0} in RADIS database'.format(state.get_fullname())))

    # Return
    return state


# %% Test


if __name__ == '__main__':
    
    from radis.test.db.test_molecules import _run_testcases
    print(('Testing molecules.py', _run_testcases()))



# -*- coding: utf-8 -*-
"""
Functions specific to the calculation of CO2 energy levels

Notes
-----

@dev: you can define Python functions here, with hardcoded spectroscopic coefficients. 
Or, you can use the spectroscopic coefficients given in ``radis/db/CO2`` under
JSON format. 

In both cases, you will need to overwrite the :meth:`~radis.db.classes.ElectronicState.Erovib`
method. See ``radis.db.molecules.py``


----------------------------------------------------------------------------------------------

"""


from __future__ import absolute_import
from radis.levels.vibrating_rotor import EvJ_uncoupled_vibrating_rotor, EvJah_uncoupled_vibrating_rotor

def EvJ_co2(v1, v2, l2, v3, J, coeff_dict, remove_ZPE=True):
    ''' Rovibrational energies.
    See :py:func:`~radis.levels.vibrating_rotor.EvJ_uncoupled_vibrating_rotor` 
    '''
    return EvJ_uncoupled_vibrating_rotor(v1, v2, l2, v3, J, coeff_dict=coeff_dict,
                                         gv1=1, gv2=2, gv3=1, remove_ZPE=remove_ZPE)


def EvJah_co2(v1, v2, l2, v3, J, coeff_dict, remove_ZPE=True):
    ''' Harmonic and anharmonic parts of rovibrational energies.
        See :py:func:`~radis.levels.vibrating_rotor.EvJah_uncoupled_vibrating_rotor` 
    '''
    return EvJah_uncoupled_vibrating_rotor(v1, v2, l2, v3, J, coeff_dict=coeff_dict,
                                         remove_ZPE=remove_ZPE)



# -*- coding: utf-8 -*-
"""
Functions and constants used in :class:`~radis.spectrum.spectrum.Spectrum` 
object

-------------------------------------------------------------------------------


"""

from __future__ import print_function, absolute_import, division, unicode_literals
from radis.misc.basics import partition

# %% Definitions

# Waverange, wavespaces
WAVENUM_UNITS = ['cm', 'cm-1', 'cm_1', 'wavenumber']
WAVELEN_UNITS = ['nm', 'wavelength']

# Spectral quantities
CONVOLUTED_QUANTITIES = ['radiance', 'transmittance', 'emissivity']
'''list: name of spectral quantities after convolution with slit function'''
NON_CONVOLUTED_QUANTITIES = ['radiance_noslit', 'transmittance_noslit',
                             'emisscoeff', 'emisscoeff_continuum', 
                             'absorbance', 'abscoeff',
                             'abscoeff_continuum', 'emissivity_noslit']
'''list: name of spectral quantities not convolved with slit function'''

# note: it is hardcoded (and needed) that quantities that are convoluted are
# generated from a non convoluted quantity with the same name + _noslit
for _q in CONVOLUTED_QUANTITIES:
    assert _q+'_noslit' in NON_CONVOLUTED_QUANTITIES

# %% Special names when printing conditions

# These parameters are shown below "Physical Conditions" rather than "Computation
# Parameters
PHYSICAL_PARAMS = ['molecule', 'wavenum_max', 'wavenum_min', 'mole_fraction',
                   'isotope', 'state', 'path_length', 'medium', 'self_absorption',
                   'slit_function_base', 'pressure_mbar',
                   'wavelength_min', 'wavelength_max',
                   'Telec', 'Tvib', 'Trot', 'Tgas', 'vib_distribution', 'rot_distribution',
                   'overpopulation', 'thermal_equilibrium']
'''list: physical conditions under which the Spectrum was calculated/measured. 
When printing an object, these parameters are shown below "Physical Conditions" 
If a parameter is not in this list, it is either in "Computation Parameters" 
(non-physical parameters that can have an influence on the Spectrum, e.g, cutoffs
and thresholds) or in "Informative Params" (descriptive parameters that have absolutely no
impact on the spectrum, e.g, number of lines calculated or calculation time)
'''

# Informative parameters
# ... Parameters that should be saved in the Spectrum objects, but ignored when
# ... comparing two spectra.
# ... should be written here only these parameters that cannot affect the
# ... physical result. In particular, all parameters relative to performance
# ... should be added here.
INFORMATIVE_PARAMS = [
    'db_use_cached',
    'lvl_use_cached',
    'chunksize',
    'calculation_time',
    'lines_calculated',
    'lines_cutoff',
    'lines_in_continuum',
    'parallel',
    'Nprocs',
    'Ngroups',
    'warning_broadening_threshold',
    'warning_linestrength_cutoff',
    'load_energies',
    'export_lines',
    'export_populations',
]
''' list: Informative parameters. Parameters that should be saved in the Spectrum 
objects, but ignored when comparing two spectra. Should be written here only 
these parameters that cannot affect the physical result. In particular, all 
parameters relative to performance should be added here.

Notes
-----

units for these parameters are stored in Spectrum.cond_units and are defined
by the generating class (ex: SpectrumFactory)'''

# %% Util functions


def cast_waveunit(unit, force_match=True):
    ''' Standardize unit formats '''
    if unit in WAVELEN_UNITS:
        return 'nm'
    elif unit in WAVENUM_UNITS:
        return 'cm-1'
    elif force_match:
        raise ValueError('Unknown wavespace unit: {0}. Should be one of {1}'.format(unit,
                                                                                    WAVELEN_UNITS+WAVENUM_UNITS))
    else:
        return unit  # dont convert


def make_up(label):
    ''' Cosmetic changes on label, before plot 


    Parameters    
    ----------

    label: str
    '''

    # Improve units
    label = label.replace('cm_1', 'cm-1')
    label = label.replace('cm-1', 'cm$^{-1}$')
    label = label.replace('m2', 'm$^2$')
    label = label.replace('m3', 'm$^3$')
    label = label.replace('I/I0', 'I/I$_\mathrm{0}$')    # transmittance unit

    # Improve text
    if not '_noslit' in label:
        # small hack: capitalize, to make a difference with non slit value
        label = label.replace('radiance', 'Radiance')     # make a small difference between no_slit and slit while plotting
        label = label.replace('transmittance', 'Transmittance')     # make a small difference between no_slit and slit while plotting
    # ... Remove _noslit
    label = label.replace('_noslit', '')
    
    return label


def print_conditions(conditions, units,
                     phys_param_list=PHYSICAL_PARAMS, info_param_list=INFORMATIVE_PARAMS):
    ''' Print all Spectrum calculation parameters 

    Parameters
    ----------

    phys_param_list: list
        These parameters are shown below "Physical Conditions" rather than "Computation
        Parameters. See :data:`~radis.spectrum.utils.PHYSICAL_PARAMS` for more 
        information. 

    info_param_list: list
        These parameters are shown below "Information" rather than "Computation 
        Parameters. See :data:`~radis.spectrum.utils.INFORMATIVE_PARAMS` for more 
        information.
        
    See Also
    --------
    
    :data:`~radis.spectrum.utils.PHYSICAL_PARAMS`, :data:`~radis.spectrum.utils.INFORMATIVE_PARAMS`

    '''

    def align(a, space=20):
        ''' fix alignement '''
        return a + ' ' * max(1, (space - len(str(a))))

    def print_param(k):
        ''' Special formatting for nicely printing conditions '''
        v_k = conditions[k]
        # Add extra arguments based on arbitrary conditions
        args = []
        if k in units:
            args.append(units[k])
        # ... fill here for other args

        # Special formatting
        try:
            if k in ['wavenum_max_calc', 'wavenum_min_calc', 'wavelength_max', 'wavelength_min',
                     'wavenum_max', 'wavenum_min']:
                v_k_str = '{0:.4f}'.format(v_k)
            elif k in ['lines_calculated', 'lines_in_continuum', 'lines_cutoff']:
                # Add comma separator for thousands
                v_k_str = '{0:,d}'.format(v_k)
            else:
                # Default to printing str
                v_k_str = '{0}'.format(v_k)
        except ValueError:
            # Default to printing str
            v_k_str = '{0}'.format(v_k)
    
        # Crop
        if len(v_k_str) > 102:   # cut if too long
            v_k_str = v_k_str[:100] + '...'
        print(' '*2, align(k), v_k_str, *args)

    phys_param, non_phys_param = partition(lambda x: x in phys_param_list,
                                           conditions)

    info_param, non_phys_param = partition(lambda x: x in info_param_list,
                                           non_phys_param)

    print('Physical Conditions')
    print('-'*40)
    for k in sorted(phys_param):
        print_param(k)

    print('Computation Parameters')
    print('-'*40)
    for k in sorted(non_phys_param):
        print_param(k)

    if len(info_param) > 0:
        print('Information')
        print('-'*40)
        for k in sorted(info_param):
            print_param(k)

    print('-'*40)

    # print gas_inp (information on each gas slab) if exists (specifically
    # for Specair output)
    if 'gas_inp' in conditions:
        try:
            for slab in conditions['gas_inp']:
                print('Slab', slab)
                slab.print_conditions()
        except:
            pass

    return None
